Package: MAPFX
Title: MAssively Parallel Flow cytometry Xplorer (MAPFX): A Toolbox for
        Analysing Data from the Massively-Parallel Cytometry
        Experiments
Version: 1.7.0
Authors@R: 
	c(person("Hsiao-Chi", "Liao", email = "chelsea.acad@gmail.com", role = c("aut", "cre"), comment = c(ORCID = "0000-0002-9586-1246")),
	  person("Agus", "Salim", role = c("ctb")), person("infinityFlow", role = c("ctb")))
Description: MAPFX is an end-to-end toolbox that pre-processes the raw data from MPC experiments (e.g., BioLegend's LEGENDScreen and BD Lyoplates assays), and further imputes the ‘missing’ infinity markers in the wells without those measurements. The pipeline starts by performing background correction on raw intensities to remove the noise from electronic baseline restoration and fluorescence compensation by adapting a normal-exponential convolution model. Unwanted technical variation, from sources such as well effects, is then removed using a log-normal model with plate, column, and row factors, after which infinity markers are imputed using the informative backbone markers as predictors. The completed dataset can then be used for clustering and other statistical analyses. Additionally, MAPFX can be used to normalise data from FFC assays as well.
Depends: R (>= 4.4.0)
License: GPL-2
Encoding: UTF-8
RoxygenNote: 7.3.1
Roxygen: list(markdown = TRUE)
biocViews: Software, FlowCytometry, CellBasedAssays, SingleCell,
        Proteomics, Clustering
Imports: flowCore, Biobase, stringr, uwot, iCellR, igraph, ggplot2,
        RColorBrewer, Rfast, ComplexHeatmap, circlize, glmnetUtils,
        e1071, xgboost, parallel, pbapply, reshape2, gtools, utils,
        stats, cowplot, methods, grDevices, graphics
Suggests: BiocStyle, knitr, rmarkdown, testthat
VignetteBuilder: knitr
NeedsCompilation: no
URL: https://github.com/HsiaoChiLiao/MAPFX
BugReports: https://github.com/HsiaoChiLiao/MAPFX/issues
git_url: https://git.bioconductor.org/packages/MAPFX
git_branch: devel
git_last_commit: e9f080c
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.23
Date/Publication: 2025-10-31
Packaged: 2025-10-31 22:41:50 UTC; biocbuild
Author: Hsiao-Chi Liao [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-9586-1246>),
  Agus Salim [ctb],
  infinityFlow [ctb]
Maintainer: Hsiao-Chi Liao <chelsea.acad@gmail.com>

# Generated by roxygen2: do not edit by hand

export(MapfxFFC)
export(MapfxMPC)
import(ggplot2)
importFrom(Biobase,exprs)
importFrom(Biobase,pData)
importFrom(ComplexHeatmap,Heatmap)
importFrom(ComplexHeatmap,draw)
importFrom(ComplexHeatmap,rowAnnotation)
importFrom(RColorBrewer,brewer.pal)
importFrom(RColorBrewer,brewer.pal.info)
importFrom(Rfast,colsums)
importFrom(Rfast,lmfit)
importFrom(circlize,colorRamp2)
importFrom(cowplot,plot_grid)
importFrom(e1071,svm)
importFrom(flowCore,keyword)
importFrom(flowCore,logicleTransform)
importFrom(flowCore,parameters)
importFrom(flowCore,read.flowSet)
importFrom(glmnetUtils,cv.glmnet)
importFrom(grDevices,dev.off)
importFrom(grDevices,jpeg)
importFrom(graphics,abline)
importFrom(gtools,combinations)
importFrom(iCellR,Rphenograph)
importFrom(igraph,membership)
importFrom(methods,is)
importFrom(parallel,clusterEvalQ)
importFrom(parallel,clusterExport)
importFrom(parallel,makeCluster)
importFrom(parallel,stopCluster)
importFrom(pbapply,pblapply)
importFrom(reshape2,melt)
importFrom(stats,"contrasts<-")
importFrom(stats,as.formula)
importFrom(stats,contr.sum)
importFrom(stats,dexp)
importFrom(stats,dnorm)
importFrom(stats,formula)
importFrom(stats,lm)
importFrom(stats,median)
importFrom(stats,model.matrix)
importFrom(stats,optim)
importFrom(stats,pexp)
importFrom(stats,pnorm)
importFrom(stats,predict)
importFrom(stats,quantile)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stringr,str_extract)
importFrom(utils,data)
importFrom(utils,getS3method)
importFrom(utils,head)
importFrom(utils,read.csv)
importFrom(utils,write.csv)
importFrom(uwot,umap)
importFrom(xgboost,"xgb.parameters<-")
importFrom(xgboost,xgb.Booster.complete)
importFrom(xgboost,xgboost)

#' @author Hsiao-Chi Liao
#' 
#' @importFrom grDevices dev.off jpeg
#' @importFrom graphics abline
#' @importFrom methods is
#' @importFrom stats as.formula contr.sum contrasts<- dexp dnorm median model.matrix optim pexp pnorm quantile sd setNames
#' @importFrom utils head read.csv
#' 
#' 
initialize <-
function(
    path_to_fcs, #the directory of the source FCS files
    path_to_output, #the directory for saving the results
    verbose=TRUE
    ){
        
    ## The paths below have to point to directories. If they do not exist the script will create them to store outputs
    path_to_intermediary <- file.path(path_to_output,"intermediary")
    path_to_graph <- file.path(path_to_output,"graph")
    path_to_downstream <- file.path(path_to_output,"downstream")

    ## create folders
    dir.create(path_to_intermediary)
    dir.create(path_to_graph)
    dir.create(path_to_downstream)
        
    paths <- c(
    input=file.path(path_to_fcs),
    intermediary=file.path(path_to_intermediary),
    graph=file.path(path_to_graph),
    downstream=file.path(path_to_downstream)
    )
    
    paths <- vapply(paths, path.expand, "path")
        
    return(list(paths=paths))
    }

#' Imputing the unmeasured well-specific markers with regression models
#'
#' This function has been designed to impute/predict the unmeasured well-specific markers with regression models.
#'
#' @param paths a vector of characters of paths to store intput, intermediary results, outputs...
#' @param chans a vector of the names of the backbone markers (MUST match to the names in the FCS file).
#' @param yvar the name of the well-specific marker in the FCS files (has been changed to "Legend" in the first function).
#' @param cores the number of cores used to perform parallel computation (default = 8L).
#' @param models.use a vector of the names of the models used for imputation (an example: `c("LM", "LASSO3", "SVM", "XGBoost")`). The length of the vector is arbitrary.
#' @param extra_args_regression_params a list of the lists of the parameters for running regression models.
#' @param prediction_events_downsampling default = NULL (not doing subsampling). How many cells per well you want to have the imputation? (must be less than or equal to a half as we won't get the prediction for cells in the training set).
#' @param impu.training logical; if FALSE (default), not impute the training set (the dataset used to train the imputation models).
#' @param plots logical; if TRUE (default), visualise the distribution of R-sq from each infinity marker.
#' 
#' @author Hsiao-Chi Liao and InfinityFlow (Becht et. al, 2021)
#'
#' @importFrom glmnetUtils cv.glmnet
#' @importFrom e1071 svm
#' @importFrom xgboost xgboost xgb.Booster.complete xgb.parameters<-
#' @importFrom stats lm predict as.formula contr.sum contrasts<- dexp dnorm median model.matrix optim pexp pnorm quantile sd setNames formula
#' @importFrom utils getS3method head read.csv
#' @importFrom parallel makeCluster clusterExport clusterEvalQ stopCluster
#' @importFrom pbapply pblapply
#' @import ggplot2
#' @importFrom cowplot plot_grid
#' @importFrom reshape2 melt
#' @importFrom gtools combinations
#' 
#' @importFrom grDevices dev.off jpeg
#' @importFrom graphics abline
#' @importFrom methods is
#'
#' @return A list of imputations
#' 
#' @details
#' This function returns the object of imputation of the unmeasured well-specific markers. In the output directory, the imputations are saved to predictions.Rds file. Visualisation of the imputation accuracy will be provided if specified.
#'
imputation_bkb.predictors <-
function(
    paths,
    chans,
    yvar="Legend",
    cores=4L,
    models.use,
    extra_args_regression_params,
    prediction_events_downsampling=NULL,
    impu.training,
    plots=TRUE
    ){
    
    #binding the variable locally to the function
    Model <- Rsq <- NULL
    
    xp <- readRDS(file = file.path(paths["intermediary"], "impu.input_log.mt.rds"))
    metadata.cell <- readRDS(file = file.path(paths["intermediary"], "fcs_metadata_df.rds"))
    params <- extra_args_regression_params
    
    ### borrowing functions from inflow pkg ###
    ## fitters (may add arbitrary parameters in the future)
    {
    polynomial_formula <- function(variables,degree){
    n <- length(variables)
    polys <- lapply(
    seq_len(degree),
    function(deg){
    res <- apply(gtools::combinations(n,deg,variables,repeats.allowed=TRUE),1,paste,collapse="*")
    res <- paste0("I(",res,")")
    }
    )
    paste(do.call(c,lapply(polys,paste,sep="+")),collapse="+")
    }
    
    ## linear fitter
    fitter_linear <- function(x = NULL, params = NULL){
    if(!is.null(x) & !is.null(params)){
    w <- x[,"train_set"]==1
    fmla <- paste0(make.names(yvar),"~",polynomial_formula(variables=chans,degree=params$degree))
    model <- lm(formula=fmla,data=as.data.frame(x[w,c(chans,yvar)]))
    pred <- predict(model,as.data.frame(x[,chans]))
    rm(list=setdiff(ls(),c("pred","model")))
    model$model <- NULL ## Trim down for slimmer objects
    model$qr$qr <- NULL ## Trim down for slimmer objects
    return(list(pred=pred,model=model))
    }
    }
    ## lasso3 fitter
    fitter_glmnet <- function(x = NULL, params = NULL){
    if(!requireNamespace("glmnetUtils", quietly = TRUE)){
    stop("Please run install.packages(\"glmnetUtils\")")
    }
    if(!is.null(x) & !is.null(params)){
    w <- x[,"train_set"] == 1
    fmla <- paste0(make.names(yvar), "~", polynomial_formula(variables = chans, degree = params$degree))
    flma <- as.formula(fmla)
    params <- params[setdiff(names(params), "degree")]
    params <- c(
    params,
    list(
    formula = fmla,
    data = as.data.frame(x[w, c(chans, yvar)]),
    use.model.frame = TRUE
    )
    )
    
    fun <- getS3method("cv.glmnet", "formula", envir = asNamespace("glmnetUtils"))
    model <- do.call(fun, params)
    model$call <- NULL ## Slimming down object
    model$glmnet.fit$call <- NULL ## Slimming down object
    attributes(model$terms)[[".Environment"]] <- NULL ## Slimming down object
    pred <- predict(model, as.data.frame(x[, chans]), s = model$lambda.min)
    
    rm(list = setdiff(ls(), c("pred", "model")))
    return(list(pred = pred, model = model))
    }
    }
    ## svm fitter
    fitter_svm <- function(x = NULL, params = NULL){
    if(!requireNamespace("e1071", quietly = TRUE)){
    stop("Please run install.packages('e1071')")
    }
    if(!is.null(x) & !is.null(params)){
    w <- x[,"train_set"]==1
    model <- do.call(function(...){e1071::svm(...,x=x[w,chans],y=x[w,yvar])},params)
    pred <- predict(model,x[,chans])
    rm(list=setdiff(ls(),c("pred","model")))
    return(list(pred=pred,model=model))
    }
    }
    ## xgboost fitter
    fitter_xgboost <- function(x = NULL, params = NULL){
    if(!requireNamespace("xgboost", quietly = TRUE)){
    stop("Please run install.packages(\"xgboost\")")
    }
    
    if(!is.null(x) & !is.null(params)){
    w <- x[,"train_set"]==1
    args <- c(list(data = x[w, chans], label = x[w, yvar], nthread = 1L, verbose = 0), params)
    model <- do.call(xgboost::xgboost, args)
    pred <- predict(model, x[, chans])
    rm(list=setdiff(ls(),c("pred","model")))
    return(list(pred=pred,model=model))
    }
    }
    }
    ## predict_wrapper
    {
    ## xp is the matrix of predictors, x is the built model
    predict_wrapper <- function(x){
    if(is(x, "lm")){
    xp <- as.data.frame(xp)
    }
    if(is(x, "cv.glmnet")){
    requireNamespace("glmnetUtils")
    xp <- as.data.frame(xp)
    return(predict(x,xp,s=x$lambda.min)[,1])
    }
    if(is(x, "xgb.Booster")){
    requireNamespace("xgboost")
    x <- xgboost::xgb.Booster.complete(x)
    xgboost::xgb.parameters(x) <- list(nthread = 1)
    }
    if(is(x, "svm")){
    requireNamespace("e1071")
    }
    res <- predict(x, xp)
    }
    }
    ## split function for selecting training data
    {
    split_matrix <- function (mat, vector, byrow = TRUE)
    {
    if (byrow & nrow(mat) != length(vector)) {
    stop("if byrow=TRUE, vector's length should have length nrow(mat)")
    }
    else if (!byrow & ncol(mat) != length(vector)) {
    !byrow & ncol(mat) != length(vector)
    stop("if byrow=FALSE, vector's length should have length ncol(mat)")
    }
    if (byrow) {
    levels  <-  split(seq_len(nrow(mat)), vector) #rows in mat should match order in vector
    res  <-  lapply(levels, function(x) {
    mat[x, , drop = FALSE]
    })
    }
    else{
    levels  <-  split(seq_len(ncol(mat)), vector)
    res  <-  lapply(levels, function(x) {
    mat[, x, drop = FALSE]
    })
    }
    res
    }
    }
    
    ## constructing the list of regression_functions
    regression_functions <- list()
    for(m in seq_len(length(models.use))){
    if(models.use[m] == "LM"){
    regression_functions[[m]] <- fitter_linear
    }else if(models.use[m] == "LASSO3"){
    regression_functions[[m]] <- fitter_glmnet
    }else if(models.use[m] == "SVM"){
    regression_functions[[m]] <- fitter_svm
    }else{
    regression_functions[[m]] <- fitter_xgboost
    }
    }
    names(regression_functions) <- models.use
    
    
    message("\tFitting regression models...")
    chans <- make.names(chans)
    yvar <- make.names(yvar)
    colnames(xp) <- make.names(colnames(xp))
    events.code <- metadata.cell$Well.lab
    
    requireNamespace("parallel")
    
    cl <- parallel::makeCluster(min(cores,length(unique(events.code))))
    
    ## random number generation
    RNGkind("L'Ecuyer-CMRG")
    
    ## OS related
    {
    if(.Platform$OS.type == "windows") {
    mc.reset.stream <- function() return(invisible(NULL))
    } else {
    mc.reset.stream <- parallel::mc.reset.stream
    }
    
    mc.reset.stream()
    
    env <- environment()
    }
    
    ## random selecting a half of the input (stratified by wells) for training
    message("\tRandomly selecting 50% of the cells in each well for model training...")
    {
    d.e <- split_matrix(xp, events.code) #list(matrix for cells from well 1, ... well 2, ...)
    d.e <- lapply( # add another column: train_set~ 0: testing, 1: training
    d.e,
    function(x){
    w <- sample(rep(c(TRUE,FALSE),times=c(floor(nrow(x)/2),nrow(x)-floor(nrow(x)/2))))
    x <- cbind(x,train_set=ifelse(w,1,0))
    x
    }
    )
    
    train_set <- matrix(
    ncol=1,
    dimnames=list(NULL,"train_set"),
    do.call(
    c,
    lapply(
    d.e,
    function(x)
    {
    x[,"train_set"]
    }
    )
    )
    )
    # train_set
    # [1,]         0
    # [2,]         1
    # [3,]         1
    }
    
    {
    clusterExport(
    cl,
    c("yvar","chans", "regression_functions", "polynomial_formula"),
    envir=env
    )
    
    clusterEvalQ(
    cl,
    {
    chans <- make.names(chans)
    yvar <- make.names(yvar)
    }
    )
    }
    
    ## fitting models
    message("\t\tFitting...")
    {
    models <- list()
    timings <- numeric()
    
    for(i in seq_along(regression_functions)){
    message("\t\t",names(regression_functions)[i],"\n\n")
    t0 <- Sys.time()
    fun <- regression_functions[[i]]
    environment(fun) <- environment() ## Fixing issue with scoping when cores = 1L
    models[[i]] <- pblapply(
    X=d.e,
    FUN=fun,
    params=params[[i]],
    cl=cl
    )
    t1 <- Sys.time()
    dt <- difftime(t1,t0,units="secs")
    message("\t",dt," seconds","\n")
    timings <- c(timings,dt)
    }
    
    names(models) <- names(regression_functions)
    stopCluster(cl)
    saveRDS(models, file=file.path(paths["intermediary"], "regression_models.Rds"))
    saveRDS(train_set, file=file.path(paths["intermediary"], "train_set.Rds"))
    saveRDS(list(timings=timings),file.path(paths["intermediary"],"timings_train.Rds"))
    }
    metadata.cell$train_set <- train_set
    saveRDS(metadata.cell, file = file.path(paths["intermediary"], "fcs_metadata_df.rds"))
    
    ## imputation
    message("\tImputing infinity (unmeasured well-specific) markers...")
    {
    xp_orig <- xp
    xp <- xp[, chans]
    
    requireNamespace("parallel")
    cl <- makeCluster(cores)
    
    ## OS related
    {
    if(.Platform$OS.type == "windows") {
    mc.reset.stream <- function() return(invisible(NULL))
    } else {
    mc.reset.stream <- mc.reset.stream
    }
    mc.reset.stream()
    
    env <- environment()
    }
    
    #impu.training = F: no prediction for training set
    if(impu.training == FALSE){
    ## sub-sampling for prediction
    message("\t\tRandomly drawing events to predict from the test set (if it's been asked)")
    {
    spl <- split(train_set[,1], events.code) #split data by wells, train_set[,1]= a 0, 1, ... vector
    spl <- lapply(
    spl,
    function(x){ #randomly select test samples
    res <- rep(FALSE,length(x))
    w <- x==0 # if the cell is in test pool, then this returns 'true' to w
    ## sample within test
    res[w][sample(seq_len(sum(w)),min(prediction_events_downsampling,sum(w)))] <- TRUE
    res
    }
    )
    
    pred_set <- which(do.call(c,spl)) #combine testing cells from different wells
    }
    
    
    xp <- xp[pred_set,]
    xp_orig <- xp_orig[pred_set, ]
    
    {
    clusterExport(
    cl,
    c("xp","chans","regression_functions","polynomial_formula"),
    envir=env
    )
    invisible(clusterEvalQ(
    cl,
    {
    colnames(xp) <- make.names(colnames(xp))
    xp <- xp[,make.names(chans)]
    }
    ))
    }
    
    for(i in seq_along(models)){
    models[[i]] <- lapply(models[[i]],"[[",2)
    }
    
    ## imputing the testing set
    message("\t\tImputing...")
    {
    preds <- list()
    timings <- numeric()
    fun <- predict_wrapper
    environment(fun) <- environment() ## Fixing issue with scoping when cores = 1L
    for(i in seq_along(models)){
    message("\t\t",names(models)[i],"\n\n")
    t0 <- Sys.time()
    preds[[i]] <- do.call( #length of list depends on types of models assigned by user
    cbind,
    pblapply(
    cl=cl,# no. core
    X=models[[i]], # used model
    FUN=fun # function do prediction
    )
    )
    t1 <- Sys.time()
    dt <- difftime(t1,t0,units="secs")
    message("\t",dt," seconds","\n")
    timings <- c(timings,dt)
    colnames(preds[[i]]) <- names(models[[i]])
    }
    stopCluster(cl)
    }
    
    
    message("\t\tConcatenating predictions...")
    {
    preds <- lapply(preds,function(x){cbind(xp_orig,x)}) #cbind(xp_orig: original in data; x: imputed variables)
    preds <- lapply(preds, as.matrix)
    names(preds) <- names(models)
    }
    
    message("\t\tWriting to disk...")
    {
    saveRDS(preds, file=file.path(paths["downstream"],"predictions.Rds"))
    saveRDS(pred_set, file=file.path(paths["intermediary"],"sampling_preds.Rds")) #should be the same as from inflow.logi.pipe
    saveRDS(list(timings=timings),file.path(paths["intermediary"],"timings_pred.Rds"))
    }
    }else{
    ##impu training
    {
    ## sub-sampling for prediction
    message("\t\tRandomly drawing events (if specified) from the training set for prediction.")
    {
    spl <- split(train_set[,1], events.code) #split data by wells, train_set[,1]= a 0, 1, ... vector
    spl <- lapply(
    spl,
    function(x){ #randomly select test samples
    res <- rep(FALSE,length(x))
    w <- x==1 # if the cell is in test pool, then this returns 'true' to w
    ## sample within test
    res[w][sample(seq_len(sum(w)),min(prediction_events_downsampling,sum(w)))] <- TRUE
    res
    }
    )
    
    pred_set <- which(do.call(c,spl)) #combine testing cells from different wells
    }
    
    
    xp <- xp[pred_set,]
    xp_orig <- xp_orig[pred_set, ]
    
    ## not sure what these used for... about parallel computing??
    {
    clusterExport(
    cl,
    c("xp","chans","regression_functions","polynomial_formula"),
    envir=env
    )
    invisible(clusterEvalQ(
    cl,
    {
    colnames(xp) <- make.names(colnames(xp))
    xp <- xp[,make.names(chans)]
    }
    ))
    }
    
    for(i in seq_along(models)){
    models[[i]] <- lapply(models[[i]],"[[",2)
    }
    
    ## imputing the testing set
    message("\t\tImputing...")
    {
    preds <- list()
    timings <- numeric()
    fun <- predict_wrapper
    environment(fun) <- environment() ## Fixing issue with scoping when cores = 1L
    for(i in seq_along(models)){
    message("\t\t",names(models)[i],"\n\n")
    t0 <- Sys.time()
    preds[[i]] <- do.call( #length of list depends on types of models assigned by user
    cbind,
    pblapply(
    cl=cl,# no. core
    X=models[[i]], # used model
    FUN=fun # function do prediction
    )
    )
    t1 <- Sys.time()
    dt <- difftime(t1,t0,units="secs")
    message("\t",dt," seconds","\n")
    timings <- c(timings,dt)
    colnames(preds[[i]]) <- names(models[[i]])
    }
    stopCluster(cl)
    }
    
    message("\t\tConcatenating predictions...")
    {
    preds <- lapply(preds,function(x){cbind(xp_orig,x)}) #cbind(xp_orig: original in data; x: imputed variables)
    preds <- lapply(preds, as.matrix)
    names(preds) <- names(models)
    }
    
    message("\t\tWriting to disk...")
    {
    saveRDS(preds, file=file.path(paths["downstream"],"predictions_training.Rds"))
    saveRDS(pred_set, file=file.path(paths["intermediary"],"sampling_preds_training.Rds")) #should be the same as from inflow.logi.pipe
    saveRDS(list(timings=timings),file.path(paths["intermediary"],"timings_pred_training.Rds"))
    }
    }
    ##impu testing
    {
    ## sub-sampling for prediction
    message("\t\tRandomly drawing events (if specified) from the testing set for prediction.")
    {
    spl <- split(train_set[,1], events.code) #split data by wells, train_set[,1]= a 0, 1, ... vector
    spl <- lapply(
    spl,
    function(x){ #randomly select test samples
    res <- rep(FALSE,length(x))
    w <- x==0 # if the cell is in test pool, then this returns 'true' to w
    ## sample within test
    res[w][sample(seq_len(sum(w)),min(prediction_events_downsampling,sum(w)))] <- TRUE
    res
    }
    )
    
    pred_set <- which(do.call(c,spl)) #combine testing cells from different wells
    }
    
    
    xp <- xp[pred_set,]
    xp_orig <- xp_orig[pred_set, ]
    
    {
    clusterExport(
    cl,
    c("xp","chans","regression_functions","polynomial_formula"),
    envir=env
    )
    invisible(clusterEvalQ(
    cl,
    {
    colnames(xp) <- make.names(colnames(xp))
    xp <- xp[,make.names(chans)]
    }
    ))
    }
    
    for(i in seq_along(models)){
    models[[i]] <- lapply(models[[i]],"[[",2)
    }
    
    ## imputing the testing set
    message("\t\tImputing...")
    {
    preds <- list()
    timings <- numeric()
    fun <- predict_wrapper
    environment(fun) <- environment() ## Fixing issue with scoping when cores = 1L
    for(i in seq_along(models)){
    message("\t\t",names(models)[i],"\n\n")
    t0 <- Sys.time()
    preds[[i]] <- do.call( #length of list depends on types of models assigned by user
    cbind,
    pblapply(
    cl=cl,# no. core
    X=models[[i]], # used model
    FUN=fun # function do prediction
    )
    )
    t1 <- Sys.time()
    dt <- difftime(t1,t0,units="secs")
    message("\t",dt," seconds","\n")
    timings <- c(timings,dt)
    colnames(preds[[i]]) <- names(models[[i]])
    }
    stopCluster(cl)
    }
    
    
    message("\t\tConcatenating predictions...")
    {
    preds <- lapply(preds,function(x){cbind(xp_orig,x)}) #cbind(xp_orig: original in data; x: imputed variables)
    preds <- lapply(preds, as.matrix)
    names(preds) <- names(models)
    }
    
    message("\t\tWriting to disk...")
    {
    saveRDS(preds, file=file.path(paths["downstream"],"predictions.Rds"))
    saveRDS(pred_set, file=file.path(paths["intermediary"],"sampling_preds.Rds")) #should be the same as from inflow.logi.pipe
    saveRDS(list(timings=timings),file.path(paths["intermediary"],"timings_pred.Rds"))
    }
    }
    }
    
    
    {
    rsq.ls <- list()
    for(i in seq_along(preds)){
    preds.here <- preds[[i]]
    dat.for.rsq <- data.frame(
    Well.lab=metadata.cell$Well.lab[which(metadata.cell$train_set == 0)],
    train_set=metadata.cell$train_set[which(metadata.cell$train_set == 0)],
    preds.here[,-match(chans, colnames(preds.here))])
    uniq.well <- unique(dat.for.rsq$Well.lab)
    rsq.v <- c()
    for(j in seq_along(uniq.well)){
    dat.here <- dat.for.rsq[which(dat.for.rsq$Well.lab == uniq.well[j]),]
    obs <- dat.here$Legend
    pred <- dat.here[,(3+j)]
    
    residual <- (obs - pred)
    
    SSE <- sum(residual^2)
    SSTO <- sum((obs - mean(obs))^2)
    
    rsq.v[j] <- (1-(SSE/SSTO))
    }
    names(rsq.v) <- uniq.well
    rsq.ls[[i]] <- rsq.v
    }
    names(rsq.ls) <- names(preds)
    saveRDS(rsq.ls, file = file.path(paths["intermediary"], "Rsq.Rds"))
    
    if(plots == TRUE){
    message("\t\tVisualising the accuracy of the predictions... (using testing set)")
    ## plotting r-sq
    rsq.df <- do.call(cbind, rsq.ls)
    {
    if(length(rsq.ls) > 1){
    graph.dat <- melt(data.frame(rsq.df))
    colnames(graph.dat) <- c("Model", "Rsq")
    }else{
    graph.dat <- data.frame(rsq.df, Model=rep(colnames(rsq.df), nrow(rsq.df)))
    colnames(graph.dat)[1] <- "Rsq"
    }
    
    if(length(rsq.ls) == 1){
    plt1 <- ggplot(data=graph.dat, mapping = aes(x=Model, y=Rsq)) +
    geom_boxplot(fill = "magenta3", color = "black") + 
    coord_flip() +
    xlab("") +
    theme(axis.text.y=element_blank(),
    axis.ticks.y=element_blank()) +
    theme_classic()
    
    plt2 <- ggplot(data=graph.dat, mapping = aes(x=Rsq)) +
    labs(title = "The Performance of Imputation",
    subtitle = paste0("Imputation Models: ", paste(names(rsq.ls), collapse = ","))) +
    geom_histogram(fill = "magenta", color = "black") +
    ylab("Counts") +
    theme_classic()
    
    jpeg(filename = file.path(paths["graph"], paste0("Rsq_boxplot_", length(rsq.ls),"model_500x800.jpeg")), height = 500, width = 800, res = 80)
    p <- plot_grid(plt2, plt1, 
    ncol = 1, rel_heights = c(2, 1),
    align = 'v', axis = 'lr')
    plot(p)
    dev.off()
    }else{ #more than one imputation method (boxplot only)
    jpeg(filename = file.path(paths["graph"], paste0("Rsq_boxplot_", length(rsq.ls),"model_500x800.jpeg")), height = 500, width = 800, res = 80)
    p <- ggplot(data=graph.dat, mapping = aes(x=Model, y=Rsq, fill=Model)) +
    geom_boxplot() +
    coord_flip() +
    labs(title = "The Performance of Imputation",
    subtitle = paste0("Imputation Models: ", paste(names(rsq.ls), collapse = ", "))) +
    theme(legend.position="none",
    plot.title = element_text(size=20),
    plot.subtitle = element_text(size=16),
    axis.text=element_text(size=14, face = "bold"),
    axis.title=element_text(size=18,face="bold"),
    )
    plot(p)
    dev.off()
    }
    }
    }
    }
    }
    return(preds)
    }
    

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo = TRUE)

## ----metaformat1, echo = FALSE, results = 'asis'------------------------------
library(knitr)
df <- data.frame(
    Filenam = c("p1_a12.fcs", "p2_d08.fcs", "p3_g1.fcs"), #should be the same as the GUID in the FCS files
    Plate = c("Plate1", "Plate2", "Plate3"),
    Well = c("A12", "D08", "G01"),
    Column = c("Col.12", "Col.08", "Col.01"),
    Row = c("Row.01", "Row.04", "Row.07"),
    Well.lab = c("P1_A12", "P2_D08", "P3_G01"))
kable(df)

## ----metaformat2, echo = FALSE, results = 'asis'------------------------------
library(knitr)
df <- data.frame(
    Filenam = c("090122.fcs", "070122.fcs", "010122.fcs"), #should be the same as the GUID in the FCS files
    Batch = c("Batch1", "Batch2", "Batch3"))
kable(df)

## ----install_mapfx, eval=FALSE------------------------------------------------
# if (!require("BiocManager", quietly = TRUE))
#     install.packages("BiocManager")
# 
# BiocManager::install("MAPFX")

## ----dependency---------------------------------------------------------------
library(MAPFX)

## specify the package names
suppressPackageStartupMessages({
library(flowCore)
library(Biobase)
library(stringr)
library(uwot)
library(iCellR)
library(igraph)
library(ggplot2)
library(RColorBrewer)
library(Rfast)
library(ComplexHeatmap)
library(circlize)
library(glmnetUtils)
library(e1071)
library(xgboost)
library(parallel)
library(pbapply)
library(reshape2)
library(gtools)
library(utils)
library(stats)
library(cowplot)
})


## ----MapfxMPC_impu_example, eval=TRUE-----------------------------------------
# import built-in data
data(ord.fcs.raw.meta.df.out_mpc)
data(ord.fcs.raw.mt_mpc)

# create an Output directory in the current working directory for the argument 'Outpath' of the MapfxMPC function
dir.create(file.path(tempdir(), "MPC_impu_Output"))

# usage
# when impute = TRUE, randomly selecting 50% of the cells in each well for model training
set.seed(123) 
MapfxMPC_impu_obj <- MapfxMPC(
    runVignette = TRUE, #set FALSE if not running this Vignette
    runVignette_meta = ord.fcs.raw.meta.df.out_mpc, #set NULL if not running this Vignette
    runVignette_rawInten = ord.fcs.raw.mt_mpc, #set NULL if not running this Vignette
    FCSpath = NULL, # users specify their own input path
    Outpath = file.path(tempdir(), "MPC_impu_Output"), # or users specify their own output path
    file_meta = "auto",
    bkb.v = c(
    "FSC-H", "FSC-W", "SSC-H", "SSC-W", "CD69-CD301b", "MHCII", 
    "CD4", "CD44", "CD8", "CD11c", "CD11b", "F480", 
    "Ly6C", "Lineage", "CD45a488", "CD24", "CD103"),
    yvar = "Legend", 
    control.wells = c(
    "P1_A01", "P2_A01", "P3_A01",
    "P3_F04", "P3_F05", "P3_F06", "P3_F07", "P3_F08", 
    "P3_F09", "P3_F10", "P3_F11", "P3_F12",
    "P3_G01", "P3_G02"),
    bkb.upper.quantile = 0.9, 
    bkb.lower.quantile = 0.1, 
    bkb.min.quantile = 0.01,
    inf.lower.quantile = 0.1, 
    inf.min.quantile = 0.01, 
    plots.bkc.bkb = TRUE, plots.bkc.inf = TRUE, 
    plots.initM = TRUE,
    plots.rmWellEffect = TRUE,
    impute = TRUE,
    models.use = c("XGBoost"),
    extra_args_regression_params = list(list(nrounds = 1500, eta = 0.03)),
    prediction_events_downsampling = NULL,
    impu.training = FALSE,
    plots.imputation = TRUE,
    cluster.analysis.bkb = TRUE, plots.cluster.analysis.bkb = TRUE,
    cluster.analysis.all = TRUE, plots.cluster.analysis.all = TRUE,
    cores = 4L)
    
# check the details
help(MapfxMPC, package = "MAPFX")

## ----mapfxNorm_MPC, eval=FALSE------------------------------------------------
# # import built-in data
# data(ord.fcs.raw.meta.df.out_mpc)
# data(ord.fcs.raw.mt_mpc)
# 
# # create an Output directory in the current working directory for the argument 'Outpath' of the MapfxMPC function
# dir.create(file.path(tempdir(), "MPC_NOimpu_Output"))
# 
# # usage
# MapfxMPC_NOimpu_obj <- MapfxMPC(
#     runVignette = TRUE, #set FALSE if not running this Vignette
#     runVignette_meta = ord.fcs.raw.meta.df.out_mpc, #set NULL if not running this Vignette
#     runVignette_rawInten = ord.fcs.raw.mt_mpc, #set NULL if not running this Vignette
#     FCSpath = NULL, # users specify their own input path
#     Outpath = file.path(tempdir(), "MPC_NOimpu_Output"), # or users specify their own output path
#     file_meta="auto",
#     bkb.v = c(
#     "FSC-H", "FSC-W", "SSC-H", "SSC-W", "CD69-CD301b", "MHCII",
#     "CD4", "CD44", "CD8", "CD11c", "CD11b", "F480",
#     "Ly6C", "Lineage", "CD45a488", "CD24", "CD103"),
#     yvar="Legend",
#     control.wells = c(
#     "P1_A01", "P2_A01", "P3_A01",
#     "P3_F04", "P3_F05", "P3_F06", "P3_F07", "P3_F08",
#     "P3_F09", "P3_F10", "P3_F11", "P3_F12",
#     "P3_G01", "P3_G02"),
#     bkb.upper.quantile = 0.9,
#     bkb.lower.quantile = 0.1,
#     bkb.min.quantile = 0.01,
#     inf.lower.quantile = 0.1,
#     inf.min.quantile = 0.01,
#     plots.bkc.bkb = TRUE, plots.bkc.inf = TRUE,
#     plots.initM = TRUE,
#     plots.rmWellEffect = TRUE,
#     impute = FALSE,
#     cluster.analysis.bkb = TRUE, plots.cluster.analysis.bkb = TRUE,
#     cores = 4L)
# 
# # check the details
# help(MapfxMPC, package = "MAPFX")

## ----mapfxNorm_FFC, eval=TRUE-------------------------------------------------
# import built-in data
data(ord.fcs.raw.meta.df.out_ffc)
data(ord.fcs.raw.mt_ffc)

# create an Output directory in the current working directory for the argument 'Outpath' of the MapfxFFC function
dir.create(file.path(tempdir(), "FFCnorm_Output"))

MapfxFFC_obj <- MapfxFFC(
    runVignette = TRUE, #set FALSE if not running this Vignette
    runVignette_meta = ord.fcs.raw.meta.df.out_ffc, #set NULL if not running this Vignette
    runVignette_rawInten = ord.fcs.raw.mt_ffc, #set NULL if not running this Vignette
    FCSpath = NULL, # users specify their own input path
    Outpath = file.path(tempdir(), "FFCnorm_Output"), # or users specify their own output path
    protein.v = c("CD3","CD4","CD8","CD45"),
    protein.upper.quantile = 0.9, 
    protein.lower.quantile = 0.1, 
    protein.min.quantile = 0.01,
    plots.bkc.protein = TRUE,
    plots.initM = TRUE,
    plots.rmBatchEffect = TRUE,
    cluster.analysis.protein = TRUE, plots.cluster.analysis.protein = TRUE)
    
# check the details
help(MapfxFFC, package = "MAPFX")

## ----sessinfo-----------------------------------------------------------------
sessionInfo()

#' Cluster analysis with normalised backbone measurements
#' 
#' This function has been designed to perform cluster analysis for the normalised backbone measurements.
#' 
#' @param paths a vector of characters of paths to store intput, intermediary results, outputs...
#' @param bkb.v a vector of the names of the backbone markers (MUST match to the names in the FCS file).
#' @param plots logical; if TRUE (default), produce an UMAP embedding plot from the normalised backbone markers to visualise the structure of the biological clusters.
#' 
#' @author Hsiao-Chi Liao
#' 
#' @import ggplot2
#' @importFrom uwot umap
#' @importFrom iCellR Rphenograph
#' @importFrom igraph membership
#' @importFrom RColorBrewer brewer.pal.info brewer.pal
#' 
#' @importFrom grDevices dev.off jpeg
#' @importFrom graphics abline
#' @importFrom methods is
#' @importFrom stats as.formula contr.sum contrasts<- dexp dnorm median model.matrix optim pexp pnorm quantile sd setNames
#' @importFrom utils head read.csv
#' 
#' @return Metadata for cells with group labels from the cluster analysis
#' 
#' @details
#' Updating the metadata for cells in the fcs_metadata_df.rds file, adding the information of the biological clusters from the clean and complete dataset, and visualising the result with the scatter plots in the output directory.
#' 
cluster.analysis.bkbOnly <-
function(
    paths, 
    bkb.v, #bkbs involved in the plotting
    plots=TRUE
    ){
    
    #binding the variable locally to the function
    UMAP1 <- UMAP2 <- Cluster <- NULL
    
    normalised.bkb <- readRDS(file = file.path(paths["downstream"], "bkc.adj.bkb_logScale_mt.rds")) 
    metadata.cell <- readRDS(file = file.path(paths["intermediary"], "fcs_metadata_df.rds"))
    names(bkb.v) <- make.names(bkb.v)
    
    message("Cluster analysis for normalised backbone measurements...")
    message("Clustering with normalised backbones")
    bkb.dat <- normalised.bkb[,match(bkb.v,colnames(normalised.bkb)),drop=FALSE]
    message("Running UMAP...")
    umap.bkb <- umap(bkb.dat, n_neighbors = 15, min_dist = 0.2, metric = "euclidean", n_epochs = 2000)
    saveRDS(umap.bkb, file = file.path(paths["downstream"], paste0("ClusterAnalysis_", "umap_",length(bkb.v),"bkb.rds")))
    
    message("Running Phenograph...")
    phenog.bkb <- Rphenograph(bkb.dat, k = 50)  #knn_fun = "hnsw", 
    saveRDS(phenog.bkb, file = file.path(paths["intermediary"], paste0("ClusterAnalysis_", "phenog_",length(bkb.v),"bkb.rds")))
    
    ##
    metadata.cell[,paste0("GP.denoised.bkb.allCells")] <- as.factor(membership(phenog.bkb[[2]]))
    
    if(plots == TRUE){
    message("\tVisualising clusters...")
    {
    ## bkb
    {
    graph.dat <- data.frame(umap.bkb, metadata.cell[paste0("GP.denoised.bkb.allCells")])
    colnames(graph.dat) <- c("UMAP1", "UMAP2", "Cluster")
    
    n <- length(unique(graph.dat[,3]))
    qual_col_pals <- brewer.pal.info[brewer.pal.info$category == 'qual',] #up to 74
    col.coeff <- unlist(mapply(brewer.pal, qual_col_pals$maxcolors, rownames(qual_col_pals)))[seq_len(n)]
    
    FN <- paste0("ClusterStructure_UMAP_", length(bkb.v), "bkb_colPhenog_600x750.jpeg")
    jpeg(filename = file.path(paths["graph"], FN), height = 600, width = 750, res = 80)
    p <- ggplot(
    data=graph.dat, mapping = aes(x=UMAP1, y=UMAP2, colour=Cluster)) + geom_point(size = 0.05, alpha = 0.25) + 
    labs(titles = paste0("Normalised ", length(bkb.v), "bkb. \n(", nrow(graph.dat), " cells)")) +
    guides(colour = guide_legend(override.aes = list(size=8))) + theme_bw() + 
    theme(text = element_text(size=22)) + scale_color_manual(values=col.coeff[seq_len(n)])
    plot(p)
    dev.off() 
    }
    }
    }
    
    ###
    saveRDS(metadata.cell, file = file.path(paths["downstream"], "fcs_metadata_df.rds"))
    
    message("\tCompleted!")
    
    return(metadata.cell)
    }

#' Converting FCS files to RDS files (for the case without exploratory markers)
#'
#' This function has been designed to convert the raw FCS files to data matrix and export to RDS files.
#'
#' @param paths a vector of characters of paths to store intput, intermediary results, outputs...
#' @param file_meta if the file names of the FCS files are in the specified format, set file_meta="auto"; otherwise set file_meta="usr" and provide "filename_meta.csv" in FCSpath.
#' @param MPC if the data is from MPC experiments, set MPC = TRUE. Setting FALSE represents data from the fluorescence flow cytometry (FFC) assay.
#'
#' @author Hsiao-Chi Liao
#'
#' @importFrom flowCore read.flowSet keyword parameters
#' @importFrom Biobase pData exprs
#' @importFrom stringr str_extract
#' 
#' @importFrom grDevices dev.off jpeg
#' @importFrom graphics abline
#' @importFrom methods is
#' @importFrom stats as.formula contr.sum contrasts<- dexp dnorm median model.matrix optim pexp pnorm quantile sd setNames
#' @importFrom utils head read.csv
#'
#' @return Raw protein intensities and the corresponding metadata from FFC experiments
#'
#' @details
#' Generating fcs_metadata_df.rds and fcs_rawInten_mt.rds files in the output directory.
#'
fcs_to_rds_bkb <-
function(
    paths, file_meta, MPC
    ){
    ## re-make large mt and consider SSC and FSC
    # import fcs.impu.raw files ###
    fs <- read.flowSet(path = file.path(paths["input"], "fcs"))
    name.desc.fs <- pData(parameters(fs[[1]]))
    name.desc.fs$desc[which(is.na(name.desc.fs$desc))] <- name.desc.fs$name[which(is.na(name.desc.fs$desc))]
    
    ####

    #### data - raw ####
    fcs.raw.mt.list <- fcs.raw.meta.list <- list()
    
    if((MPC == FALSE) & (file_meta == "usr")){
    meta.inf <- read.csv(file = file.path(paths["input"], "meta", "filename_meta.csv"))
    for(i in seq_along(fs)){
    fcs.raw.mt.list[[i]] <- exprs(fs[[i]])
    filenam <- unlist(keyword(fs[[i]], "GUID"))
    fcs.raw.meta.list[[i]] <- cbind(
    NO.in.batch=(seq_len(dim(fcs.raw.mt.list[[i]])[1])),
    Filenam=filenam,
    Batch=meta.inf[which(filenam == meta.inf$Filenam), "Batch"])
    }
    fcs.raw.mt <- do.call(rbind, fcs.raw.mt.list)
    colnames(fcs.raw.mt) <- name.desc.fs$desc[match(colnames(fcs.raw.mt), name.desc.fs$name)]
    ####
    
    # making metadata ###
    fcs.raw.meta.df <- as.data.frame(do.call(rbind, fcs.raw.meta.list))
    fcs.raw.meta.df$NO.in.batch <- as.integer(fcs.raw.meta.df$NO.in.batch)
    #use lapply for multiple changing class: lapply(mydf[,2:3], as.factor)
    
    #ordering before giving NO.in.all
    ord.fcs.raw.meta.df <- fcs.raw.meta.df[order(fcs.raw.meta.df$Batch),,drop=FALSE]
    ord.fcs.raw.mt <- fcs.raw.mt[order(fcs.raw.meta.df$Batch),,drop=FALSE]
    
    #global ID for each cell
    ord.fcs.raw.meta.df$NO.in.all <- seq_len(nrow(ord.fcs.raw.mt))
    
    #out rds
    ord.fcs.raw.meta.df.out <- ord.fcs.raw.meta.df[,c(4,1,3),drop=FALSE]
    }else{
    stop("Please provide a \"filename_meta.csv\" file in `FCSpath/meta/`")
    }
    
    saveRDS(ord.fcs.raw.meta.df.out, file = file.path(paths["intermediary"], "fcs_metadata_df.rds"))
    saveRDS(ord.fcs.raw.mt, file = file.path(paths["intermediary"], "fcs_rawInten_mt.rds"))
    }

#' MAssively Parallel Flow cytometry Xplorer (MAPFX)
#'
#' @description This function is an end-to-end toolbox for analysing single-cell protein intensity data from the Massively-Parallel Cytometry (MPC) Experiments in FCS format. The functions include data normalisation, imputation (using backbone markers), and cluster analysis.
#'
#' @param runVignette logical; if FALSE (default), specify a path to `FCSpath` argument; TRUE for running Vignette using built-in data. 
#' @param runVignette_meta the argument for the built-in metadata when running Vignette; NULL (default).
#' @param runVignette_rawInten the argument for the built-in raw intensities when running Vignette; NULL (default).
#' @param FCSpath path to the input directory where `filename_meta.csv` and FCS files are stored (one file per well). `filename_meta.csv` should be saved under `FCSpath/FCS/meta/` and FCS files should be saved under `FCSpath/FCS/fcs/` (See Vignette for details.)
#' @param Outpath path to the output directory where intermediate results and final results will be stored.
#' @param file_meta if the file names of the FCS files are in the specified format, set `file_meta = "auto"`; otherwise set `file_meta = "usr"` and provide a `filename_meta.csv` file in `FCSpath/FCS/meta/`.
#' @param bkb.v a vector of the names of the backbone markers (MUST be the same as the names in the FCS files). For example, `bkb.v = c("FSC-H", "FSC-W", "SSC-H", "SSC-W", "CD69-CD301b", "MHCII", "CD4", "CD44", "CD8", "CD11c", "CD11b", "F480", "Ly6C", "Lineage", "CD45a488", "CD24", "CD103")`.
#' @param yvar the name of the well-specific exploratory marker in the FCS files (e.g., "Legend").
#' @param control.wells the well label of the control wells, including the autofluorescence and the isotype controls (format: plate_well, e.g., P1_A01). Users need to provide this information when `cluster.analysis.all = TRUE`. For example, `control.wells = c("P1_A01", "P2_A01", "P3_A01", "P3_F04", "P3_F05", "P3_F06", "P3_F07", "P3_F08", "P3_F09", "P3_F10", "P3_F11", "P3_F12", "P3_G01", "P3_G02")`.
#' @param bkb.upper.quantile the cut-off (default = 0.9) for selecting cells used for estimating the parameter of signal for backbone markers.
#' @param bkb.lower.quantile the cut-off (default = 0.1) for selecting cells used for estimating the parameters of noise for backbone markers.
#' @param bkb.min.quantile the cut-off (default = 0.01) for omitting the cells with the smallest values to minimise the impact of outliers during estimation (backbone).
#' @param inf.lower.quantile the cut-off (default = 0.1) for selecting cells used for estimating the parameters of noise for infinity markers.
#' @param inf.min.quantile the cut-off (default = 0.01) for omitting the cells with the smallest values to minimise the impact of outliers during estimation (infinity).
#' @param plots.bkc.bkb logical; if TRUE (default), produce scatter plots for pre- and post- background adjusted backbone markers (calibrated values on y-axis and raw values on x-axis).
#' @param plots.bkc.inf logical; if TRUE (default), produce scatter plots for pre- and post- background adjusted infinity markers (calibrated values on y-axis and raw values on x-axis).
#' @param plots.initM logical; if TRUE (default), produce an UMAP embedding plot to visualise the structure of the biological clusters used to form the initial M matrix for removal of well effects.
#' @param plots.rmWellEffect logical; if TRUE (default), produce heatmaps to visualise the unwanted (well) effects and biological effects in the pre- and post- adjusted datasets.
#' @param impute logical; if TRUE (default), impute the missing infinity markers.
#' @param models.use a vector of the names of the models used for imputation. For example, `models.use = c("LM", "LASSO3", "SVM", "XGBoost")`.
#' @param extra_args_regression_params a list of the lists of the parameters for running regression models. The order should be the same as the models specified in `models.use`. For example, `extra_args_regression_params = list(list(degree = 1), list(nfolds = 10, degree = 3), list(type = "nu-regression", cost = 8, nu = 0.5, kernel = "radial"), list(nrounds = 1500, eta = 0.03))`.
#' @param prediction_events_downsampling integer (default = NULL); the number of samples used for the downsampling for the prediction.
#' @param impu.training logical; if FALSE (default), not impute the training set (the dataset used to train the imputation models).
#' @param plots.imputation logical; if TRUE (default), visualise the distribution of R-sq values of infinity markers.
#' @param cluster.analysis.bkb  logical; if TRUE (default), perform cluster analysis using normalised backbone markers for all cells.
#' @param plots.cluster.analysis.bkb logical; if TRUE (default), produce an UMAP embedding plot from the normalised backbone markers to visualise the structure of the biological clusters for all cells.
#' @param cluster.analysis.all logical; must set `FALSE` if `impute = FALSE`; if TRUE (default), perform cluster analysis using normalised backbone markers and imputed infinity markers for cells in testing set.
#' @param plots.cluster.analysis.all logical; must set `FALSE` if `impute = FALSE`; if TRUE (default), produce an UMAP embedding plot from the normalised backbone markers and the imputed infinity markers to visualise the structure of the biological clusters for cells in testing set.
#' @param cores the number of cores used to perform parallel computation during the imputation process (default = 4L).
#'
#' @author Hsiao-Chi Liao, Agus Salim, and InfinityFlow (Becht et. al, 2021)
#'
#' @importFrom utils data
#'
#' @export
#' 
#' @return Normalised backbone markers on log scale, background noise corrected infinity markers, imputations, and metadata for cells
#' 
#' @details
#' In the output directory, this function produces the normalised backbone measurements, the background corrected infinity measurements, and imputed infinity markers (if set impute = TRUE), cell group labels from the cluster analysis using both normalised backbones and the completed dataset (if impute = TRUE), and graphs will be provided if specified.
#' 
#' @examples
#' # import built-in data
#' data(ord.fcs.raw.meta.df.out_mpc)
#' data(ord.fcs.raw.mt_mpc)
#' 
#' # create an Output directory for the MapfxMPC function
#' dir.create(file.path(tempdir(), "MPC_impu_Output"))
#' 
#' # When `impute = TRUE`, randomly selecting 50% of the cells in each well for model training
#' set.seed(123) 
#' MapfxMPC_impu_obj <- MapfxMPC(
#'   runVignette = TRUE, #set FALSE if not running this example
#'   runVignette_meta = ord.fcs.raw.meta.df.out_mpc, #set NULL if not running this example
#'   runVignette_rawInten = ord.fcs.raw.mt_mpc, #set NULL if not running this example
#'   FCSpath = NULL, # users specify their own input path if not running this example
#'   Outpath = file.path(tempdir(), "MPC_impu_Output"),
#'   file_meta = "auto",
#'   bkb.v = c(
#'     "FSC-H", "FSC-W", "SSC-H", "SSC-W", "CD69-CD301b", "MHCII", 
#'     "CD4", "CD44", "CD8", "CD11c", "CD11b", "F480", 
#'     "Ly6C", "Lineage", "CD45a488", "CD24", "CD103"),
#'   yvar = "Legend", 
#'   control.wells = c(
#'     "P1_A01", "P2_A01", "P3_A01",
#'     "P3_F04", "P3_F05", "P3_F06", "P3_F07", "P3_F08", 
#'     "P3_F09", "P3_F10", "P3_F11", "P3_F12",
#'     "P3_G01", "P3_G02"),
#'   bkb.upper.quantile = 0.9, 
#'   bkb.lower.quantile = 0.1, 
#'   bkb.min.quantile = 0.01,
#'   inf.lower.quantile = 0.1, 
#'   inf.min.quantile = 0.01, 
#'   plots.bkc.bkb = TRUE, plots.bkc.inf = TRUE, 
#'   plots.initM = TRUE,
#'   plots.rmWellEffect = TRUE,
#'   impute = TRUE,
#'   models.use = c("XGBoost"),
#'   extra_args_regression_params = list(list(nrounds = 1500, eta = 0.03)),
#'   prediction_events_downsampling = NULL,
#'   impu.training = FALSE,
#'   plots.imputation = TRUE,
#'   cluster.analysis.bkb = TRUE, plots.cluster.analysis.bkb = TRUE,
#'   cluster.analysis.all = TRUE, plots.cluster.analysis.all = TRUE,
#'   cores = 2L)
#' 
#' 
MapfxMPC <-
function(
    runVignette = FALSE,
    runVignette_meta = NULL,
    runVignette_rawInten = NULL,
    FCSpath = NULL,
    Outpath = NULL,
    file_meta = "auto",
    bkb.v = NULL,
    yvar = "Legend",
    control.wells = NULL,
    bkb.upper.quantile = 0.9, 
    bkb.lower.quantile = 0.1, 
    bkb.min.quantile = 0.01,
    inf.lower.quantile = 0.1, 
    inf.min.quantile = 0.01, 
    plots.bkc.bkb = TRUE, plots.bkc.inf = TRUE, 
    plots.initM = TRUE,
    plots.rmWellEffect = TRUE,
    impute = TRUE,
    models.use = c("XGBoost"),
    extra_args_regression_params = list(list(nrounds = 1500, eta = 0.03)),
    prediction_events_downsampling = NULL,
    impu.training = FALSE,
    plots.imputation = TRUE,
    cluster.analysis.bkb = TRUE, plots.cluster.analysis.bkb = TRUE,
    cluster.analysis.all = TRUE, plots.cluster.analysis.all = TRUE,
    cores = 4L){
    
    chans <- bkb.v
    
    #run Vignette
    if(runVignette == TRUE){
    if(is.null(Outpath)){
    message("\nPlease make sure the output path has been specified!")
    }
    
    message("\n\n\nCreating directories for output...")
    settings <- initialize(
    path_to_fcs=FCSpath,
    path_to_output=Outpath,
    verbose=TRUE)
    paths <- settings$paths
    
    ###
    
    saveRDS(runVignette_meta, file = file.path(paths["intermediary"], "fcs_metadata_df.rds"))
    saveRDS(runVignette_rawInten, file = file.path(paths["intermediary"], "fcs_rawInten_mt.rds"))
    }
    
    #not running Vignette
    if(runVignette == FALSE){
    
    if(is.null(FCSpath) | is.null(Outpath) | is.null(bkb.v)){
    stop("\nPlease make sure you have specified\n(1) path to the input files (details in the folder diagram in the vignette) \n(2) output path \n(3) a list of backbone markers")
    }
    
    if(is.null(file_meta)){
    stop("\nThe file_meta argument only accepts either \"auto\" or \"usr\".\nAlso, if file_meta=\"usr\", make sure there is a filename_meta.csv file in FCSpath/meta (details in the folder diagram in the vignette).")
    }else if(file_meta != "auto" & file_meta != "usr"){
    stop("\nThe file_meta argument only accepts either \"auto\" or \"usr\".\nAlso, if file_meta=\"usr\", make sure there is a filename_meta.csv file in FCSpath/meta (details in the folder diagram in the vignette).")
    }
    
    if(!is.null(FCSpath) & !is.null(Outpath) & !is.null(bkb.v)){
    
    if(cluster.analysis.all == TRUE & is.null(control.wells)){
    stop("Please specify your control wells because they will be removed for cluster analysis...")
    }
    
    if(length(extra_args_regression_params) != length(models.use)){
    stop("'extra_args_regression_params' and 'models.use' should be lists of the same lengths")
    }
    
    ##/!\ Potentially add a check here to make sure parameters are consistent with FCS files
    message("\n\n\nCreating directories for output...")
    
    settings <- initialize(
    path_to_fcs=FCSpath,
    path_to_output=Outpath,
    verbose=TRUE)
    paths <- settings$paths
    
    ###
    
    ##1.fcs to rds
    message("\n\n\nTransforming FCS to RDS files...")
    fcs_to_rds(paths=paths, file_meta=file_meta, yvar = yvar)
    }
    }
    
    ##2-1.bkc bkb
    message("\n\n\nBackground correcting backbone markers...")
    bkc_bkb(
    paths=paths, bkb.v=bkb.v,
    MPC=TRUE,
    bkb.upper.quantile=bkb.upper.quantile, #cells used for estimating parameter of signal
    bkb.lower.quantile=bkb.lower.quantile, #cells used for estimating parameters of noise
    bkb.min.quantile=bkb.min.quantile,
    plots=plots.bkc.bkb) #the lowest 1% of values will not be used to minimise the impact of outliers on sig
    
    ##2-2.bkc inf
    message("\n\n\nBackground correcting infinity markers...")
    bkc_inf <- bkc_pe(
    paths=paths,
    pe.lower.quantile=inf.lower.quantile, #cells used for estimating parameters of noise
    pe.min.quantile=inf.min.quantile, #the lowest 1% of values will not be used to minimize the impact of outliers on sig
    plots=plots.bkc.inf)
    
    ##3-0.initM
    message("\n\n\nForming a matrix of biology (M) for removal of well effect...")
    metadata <- initM(
    paths=paths, assay="MPC", bkb.v=bkb.v,
    plots=plots.initM) #may give "cent.log.bkc" in the next version
    
    ##3-1.rmWellEffect
    message("\n\n\nRemoval of well effect...")
    norm.bkb <- rmWellEffect(
    paths=paths,
    plots = plots.rmWellEffect)
    
    if(impute == TRUE){
    message("\n\n\nImputation got started...")
    ##4.prep. for imputation - combine normalised.bkb.inf
    mkImputeMT(paths=paths)
    
    ##5.impu. using bkb as predictors
    impu <- imputation_bkb.predictors(
    paths=paths,
    chans=chans,
    yvar="Legend", #have named this in my fcs_to_rds function
    cores=cores,
    models.use = models.use,
    extra_args_regression_params = extra_args_regression_params,
    prediction_events_downsampling = prediction_events_downsampling,
    impu.training = impu.training,
    plots = plots.imputation)
    
    imputation <- impu
    
    ##
    ##6.cluster analysis for completed dataset (bkb, bkb+impu.inf) using cells from testing set
    message("\n\n\nCluster analysis with adjusted backbone markers and completed dataset for cells in the testing set...")
    if(cluster.analysis.all == TRUE){
    metadata <- cluster.analysis(
    paths=paths,
    bkb.v=bkb.v,
    yvar="Legend",
    control.wells=control.wells,
    plots = plots.cluster.analysis.all)
    message("\nCell group labels are saved in \"GP.denoised.bkb\" and \"GP.denoised.bkb.impuInf*\" columns...")
    }
    
    }else{
    imputation <- NA
    }

    ##7.cluster analysis bkbOnly using all cells
    message("\n\n\nCluster analysis with adjusted backbone markers for ALL cells...")
    if(cluster.analysis.bkb == TRUE){
    metadata <- cluster.analysis.bkbOnly(
    paths=paths,
    bkb.v=bkb.v,
    plots = plots.cluster.analysis.bkb)
    }
    message("\nCell group labels are saved in \"GP.denoised.bkb.allCells\" column...")
    
    message("\tCompleted!")
    return(list(logNormalised_bkb=norm.bkb, BKC_inf=bkc_inf, Imputation=imputation, Metadata=metadata))
    }

# Checks MapfxMPC
data(ord.fcs.raw.meta.df.out_mpc)
data(ord.fcs.raw.mt_mpc)
# create an Output directory in the current working directory for the argument 'Outpath' of the MapfxMPC function
dir.create(file.path(tempdir(), "MPC_impu_Output"))

# When `impute = TRUE`, randomly selecting 50% of the cells in each well for model training - has been tested with the example of the function
# When `impute = FALSE`, no imputation performed
set.seed(123)
MapfxMPC_NOimpu_obj <- MapfxMPC(
  runVignette = TRUE, #set FALSE if not running this example
  runVignette_meta = ord.fcs.raw.meta.df.out_mpc, #set NULL if not running this example
  runVignette_rawInten = ord.fcs.raw.mt_mpc, #set NULL if not running this example
  FCSpath = NULL, # users specify their own input path if not running this example
  Outpath = file.path(tempdir(), "MPC_impu_Output"),
  file_meta = "auto",
  bkb.v = c(
    "FSC-H", "FSC-W", "SSC-H", "SSC-W", "CD69-CD301b", "MHCII", 
    "CD4", "CD44", "CD8", "CD11c", "CD11b", "F480", 
    "Ly6C", "Lineage", "CD45a488", "CD24", "CD103"),
  yvar = "Legend", 
  control.wells = c(
    "P1_A01", "P2_A01", "P3_A01",
    "P3_F04", "P3_F05", "P3_F06", "P3_F07", "P3_F08", 
    "P3_F09", "P3_F10", "P3_F11", "P3_F12",
    "P3_G01", "P3_G02"),
  bkb.upper.quantile = 0.9, 
  bkb.lower.quantile = 0.1, 
  bkb.min.quantile = 0.01,
  inf.lower.quantile = 0.1, 
  inf.min.quantile = 0.01, 
  plots.bkc.bkb = FALSE, plots.bkc.inf = FALSE, 
  plots.initM = FALSE,
  plots.rmWellEffect = FALSE,
  impute = FALSE,
  models.use = c("XGBoost"),
  extra_args_regression_params = list(list(nrounds = 1500, eta = 0.03)),
  prediction_events_downsampling = NULL,
  impu.training = FALSE,
  plots.imputation = FALSE,
  cluster.analysis.bkb = FALSE, plots.cluster.analysis.bkb = FALSE,
  cluster.analysis.all = FALSE, plots.cluster.analysis.all = FALSE,
  cores = 2L)

test_that("MapfxMPC returns an object", {
  obj <- MapfxMPC_NOimpu_obj
  expect_true(validObject(obj))
})

test_that("MapfxMPC returns a list", {
  obj <- MapfxMPC_NOimpu_obj
  expect_true(inherits(obj, "list"))
})

test_that("MapfxMPC returns a list with size 4", {
  obj <- MapfxMPC_NOimpu_obj
  expect_length(obj, 4)
})

test_that("MapfxMPC(...,impute = FALSE) does not return imputations", {
  obj <- MapfxMPC_NOimpu_obj
  expect_equal(obj$Imputation, NA)
})

#' Cluster analysis with normalised backbone measurements and the complete dataset
#' 
#' This function has been designed to perform cluster analysis for the normalised backbone measurements and the complete dataset which includes the normalised backbone measurements and the imputed well-specific markers.
#' 
#' @param paths a vector of characters of paths to store intput, intermediary results, outputs...
#' @param bkb.v a vector of the names of the backbone markers (MUST match to the names in the FCS file).
#' @param yvar the name of the well-specific marker in the FCS files (e.g., "Legend").
#' @param control.wells the well label of the control wells, including the autofluorescence and the isotype controls (format: plate_well, e.g., P1_A01)
#' @param plots logical; if TRUE (default), produce an UMAP embedding plot from the normalised backbone markers and the imputed infinity markers to visualise the structure of the biological clusters.
#' 
#' @author Hsiao-Chi Liao
#' 
#' @import ggplot2
#' @importFrom uwot umap
#' @importFrom iCellR Rphenograph
#' @importFrom igraph membership
#' @importFrom RColorBrewer brewer.pal.info brewer.pal
#' 
#' @importFrom grDevices dev.off jpeg
#' @importFrom graphics abline
#' @importFrom methods is
#' @importFrom stats as.formula contr.sum contrasts<- dexp dnorm median model.matrix optim pexp pnorm quantile sd setNames
#' @importFrom utils head read.csv
#' 
#' @return Metadata for cells with group labels from the cluster analysis
#' 
#' @details
#' Updating the metadata for cells in the fcs_metadata_df.rds file, adding the information of the biological clusters from the clean and complete dataset, and visualising the result with the scatter plots in the output directory.
#' 
cluster.analysis <-
function(
    paths,
    bkb.v,
    yvar="Legend",
    control.wells,
    plots=TRUE
    ){
    
    #binding the variable locally to the function
    UMAP1 <- UMAP2 <- Cluster <- NULL
    
    preds <- readRDS(file = file.path(paths["downstream"], "predictions.Rds"))
    metadata.cell <- readRDS(file = file.path(paths["intermediary"], "fcs_metadata_df.rds"))
    names(bkb.v) <- make.names(bkb.v)
    
    for(i in seq_along(preds)){
    impu.mt <- preds[[i]]
    
    if(i == 1){ ## bkb only need to do once
    message("Clustering with normalised backbones")
    bkb.dat <- impu.mt[, match(names(bkb.v), colnames(impu.mt)),drop=FALSE]
    message("Running UMAP...")
    umap.bkb <- umap(
    bkb.dat,
    n_neighbors = 15, min_dist = 0.2, metric = "euclidean", n_epochs = 2000)
    saveRDS(umap.bkb, file = file.path(paths["downstream"], paste0("ClusterAnalysis_", "umap_",length(bkb.v),"bkb.rds")))
    
    message("Running Phenograph...")
    phenog.bkb <- Rphenograph(bkb.dat, k = 50)  #knn_fun = "hnsw", 
    saveRDS(phenog.bkb, file = file.path(paths["intermediary"], paste0("ClusterAnalysis_", "phenog_",length(bkb.v),"bkb.rds")))
    }
    
    ###
    message("Clustering with normalised backbones + imputed infinity markers (", names(preds)[i], ")") #now: use them all - may just pick good ones in the future
    complete.dat <- impu.mt[, -match(c(yvar, control.wells), colnames(impu.mt)),drop=FALSE]
    message("Running UMAP...")
    umap.bkb.impuInf <- umap(
    complete.dat,
    n_neighbors = 15, min_dist = 0.2, metric = "euclidean", n_epochs = 2000)
    saveRDS(umap.bkb.impuInf, 
    file = file.path(paths["downstream"], paste0("ClusterAnalysis_", names(preds)[i], "_umap_", 
    length(bkb.v),"bkb.", (ncol(complete.dat)-length(bkb.v)),"impuInf.rds")))
    
    message("Running Phenograph...")
    phenog.bkb.impuInf <- Rphenograph(complete.dat, k = 50)  #knn_fun = "hnsw", 
    saveRDS(phenog.bkb.impuInf, 
    file = file.path(paths["intermediary"], paste0("ClusterAnalysis_", names(preds)[i], "_phenog_",
    length(bkb.v),"bkb.", (ncol(complete.dat)-length(bkb.v)),"impuInf.rds")))
    
    if(i == 1){
    metadata.cell[which(metadata.cell$train_set == 0),paste0("GP.denoised.bkb")] <- as.factor(membership(phenog.bkb[[2]]))
    }
    metadata.cell[which(metadata.cell$train_set == 0),paste0("GP.denoised.bkb.impuInf_", names(preds)[i])] <- as.factor(membership(phenog.bkb.impuInf[[2]]))
    
    if(plots == TRUE){
    message("\tVisualising clusters...")
    {
    ## bkb
    {
    if(i == 1){ ## would not be affected by imputation method
    graph.dat <- data.frame(umap.bkb, metadata.cell[which(metadata.cell$train_set == 0), paste0("GP.denoised.bkb")])
    colnames(graph.dat) <- c("UMAP1", "UMAP2", "Cluster")
    
    n <- length(unique(graph.dat[,3]))
    qual_col_pals <- brewer.pal.info[brewer.pal.info$category == 'qual',] #up to 74
    col.coeff <- unlist(mapply(brewer.pal, qual_col_pals$maxcolors, rownames(qual_col_pals)))[seq_len(n)]
    
    FN <- paste0("ClusterStructure_UMAP_", length(bkb.v), "bkb_colPhenog_600x750.jpeg")
    jpeg(filename = file.path(paths["graph"], FN), height = 600, width = 750, res = 80)
    p <- ggplot(
    data=graph.dat, mapping = aes(x=UMAP1, y=UMAP2, colour=Cluster)) + geom_point(size = 0.05, alpha = 0.25) + 
    labs(titles = paste0("Normalised ", length(bkb.v), "bkb. \n(", nrow(graph.dat), " cells)")) +
    guides(colour = guide_legend(override.aes = list(size=8))) + theme_bw() + 
    theme(text = element_text(size=22)) + scale_color_manual(values=col.coeff[seq_len(n)])
    plot(p)
    dev.off() 
    }
    }
    ## bkb+impuInf
    {
    graph.dat <- data.frame(umap.bkb.impuInf, metadata.cell[which(metadata.cell$train_set == 0), paste0("GP.denoised.bkb.impuInf_",names(preds)[i])])
    colnames(graph.dat) <- c("UMAP1", "UMAP2", "Cluster")
    
    n <- length(unique(graph.dat[,3]))
    qual_col_pals <- brewer.pal.info[brewer.pal.info$category == 'qual',]
    col.coeff <- unlist(mapply(brewer.pal, qual_col_pals$maxcolors, rownames(qual_col_pals)))[seq_len(n)]
    
    FN <- paste0("ClusterStructure_UMAP_", names(preds)[i], length(bkb.v), "bkb",
    (ncol(complete.dat)-length(bkb.v)),"impuInf_colPhenog_600x750.jpeg")
    jpeg(filename = file.path(paths["graph"], FN), height = 600, width = 750, res = 80)
    p <- ggplot(
    data=graph.dat, mapping = aes(x=UMAP1, y=UMAP2, colour=Cluster)) + geom_point(size = 0.05, alpha = 0.25) + 
    labs(titles = paste0(names(preds)[i], " ", length(bkb.v), "bkb. ", (ncol(complete.dat)-length(bkb.v)),"impuInf \n(", nrow(graph.dat), " cells)")) +
    guides(colour = guide_legend(override.aes = list(size=8))) + theme_bw() + 
    theme(text = element_text(size=22)) + scale_color_manual(values=col.coeff[seq_len(n)])
    plot(p)
    dev.off()
    }
    }
    }
    }
    ###
    saveRDS(metadata.cell, file = file.path(paths["downstream"], "fcs_metadata_df.rds"))
    
    message("\tCompleted!")
    
    return(metadata.cell)
    }
    

## Documentation of data objects

#' Subset of the single-cell murine lung data at steady state from an MPC experiment
#' @docType data
#' @format a matrix containing cells from 266 wells (50 cells/well)
#' @source \url{https://flowrepository.org/id/FR-FCM-Z2LP}
#' @usage data(ord.fcs.raw.mt_ffc)
"ord.fcs.raw.mt_ffc"

#' Metadata of the example MPC data
#' @docType data
#' @format a data.frame
#' @usage data(ord.fcs.raw.meta.df.out_ffc)
"ord.fcs.raw.meta.df.out_ffc"

#' Subset of the sorted CD4+ and CD8+ T cells from mice splenocytes from an FFC experiment
#' @docType data
#' @format a matrix containing cells from 5 batches (50 cells/batch)
#' @source \url{http://flowrepository.org/id/FR-FCM-Z6UG}
#' @usage data(ord.fcs.raw.mt_mpc) 
"ord.fcs.raw.mt_mpc"

#' Metadata of the example FFC data
#' @docType data
#' @format a data.frame
#' @usage data(ord.fcs.raw.meta.df.out_mpc) 
"ord.fcs.raw.meta.df.out_mpc"

#' Initial biological clusters
#' 
#' Generating the M matrix for removing well effects.
#' 
#' This function has been designed to find initial biological clusters with centred transformed data.
#' 
#' @param paths a vector of characters of paths to store intput, intermediary results, outputs...
#' @param assay the type of the input data - MPC or FFC.
#' @param bkb.v a vector of the names of the backbone markers (MUST match to the names in the FCS file).
#' @param plots logical; if TRUE (default), produce a UMAP embedding plot to visualise the structure of the biological clusters to form the initial M matrix for removal of well effect.
#' 
#' @author Hsiao-Chi Liao
#' 
#' @import ggplot2
#' @importFrom flowCore logicleTransform
#' @importFrom uwot umap
#' @importFrom iCellR Rphenograph
#' @importFrom igraph membership
#' @importFrom RColorBrewer brewer.pal.info brewer.pal
#' 
#' @importFrom grDevices dev.off jpeg
#' @importFrom graphics abline
#' @importFrom methods is
#' @importFrom stats as.formula contr.sum contrasts<- dexp dnorm median model.matrix optim pexp pnorm quantile sd setNames
#' @importFrom utils head read.csv
#' 
#' @return Metadata for cells with the initial biological clusters labels added
#' 
#' @details
#' Updating the metadata for cells in the fcs_metadata_df.rds file, adding the information of the initial biological clusters, and visualising the result with the scatter plots in the output directory.
#' 
initM <-
function(
    paths,
    assay,
    bkb.v,
    plots=TRUE
    ){
    
    #binding the variable locally to the function
    UMAP1 <- UMAP2 <- Pheno.gp <- NULL
    
    trans.dat <- "cent.lgc" #may give "cent.log.bkc" in next version
    
    rawInten <- readRDS(file = file.path(paths["intermediary"], "fcs_rawInten_mt.rds"))
    metadata.cell <- readRDS(file = file.path(paths["intermediary"], "fcs_metadata_df.rds"))
    sel.bkb.raw <- rawInten[,match(bkb.v, colnames(rawInten)),drop=FALSE]
    
    if(trans.dat == "cent.lgc"){
    ## lgc trans - borrowing inflow people
    {
    ##bkb only
    chans <- bkb.v
    xp <- sel.bkb.raw
    message("\tForming logicle functions...")
    {
    transforms_chan <- setNames(
    lapply(
    chans,
    function(x){
    data <- xp[,x] ## all cells (together do transformation)
    
    t <- max(data)
    m <- 4.5
    
    q <- 0.05
    r <- .Machine$double.eps + quantile(data, q)
    w <- max((m-log10(t/abs(r)))/2,0.1)
    w <- min(w,m/2)
    
    a <- 0
    logicleTransform(w=w,t=t,m=m,a=a) ##Just use summary() to retrive the parameters
    }
    ),
    chans
    )
    saveRDS(transforms_chan, file=file.path(paths["intermediary"],"transform.funcs_bkb.Rds")) 
    }
    
    message("\tLogicle transforming raw intensity...")
    {
    sel.bkb.lgc <- sel.bkb.raw
    for(chan in chans){
    sel.bkb.lgc[,chan] <- transforms_chan[[chan]](xp[,chan])
    }
    }
    
    message("\tCentring logicle transformed intensities...")
    if(assay=="MPC"){
    ## MPC (centring within well)
    ## shifted by well-wise mean
    well.v <- unique(metadata.cell$Well.lab)
    shifted.ls <- list()
    for(i in seq_along(well.v)){
    dat.here <- sel.bkb.lgc[which(metadata.cell$Well.lab == well.v[i]),,drop=FALSE]
    dat.meanSF <- apply(dat.here, 2, function(x) (x-mean(x)))
    shifted.ls[[i]] <- dat.meanSF
    }
    
    sel.bkb.cent.lgc <- do.call(rbind, shifted.ls)
    saveRDS(sel.bkb.cent.lgc, file = file.path(paths["intermediary"], "sel.bkb.cent.lgc.rds"))
    }
    
    if(assay=="FFC"){
    ## FFC (centring within batch)
    batch.v <- unique(metadata.cell$Batch)
    shifted.ls <- list()
    for(i in seq_along(batch.v)){
    dat.here <- sel.bkb.lgc[which(metadata.cell$Batch == batch.v[i]),,drop=FALSE]
    dat.meanSF <- apply(dat.here, 2, function(x) (x-mean(x)))
    shifted.ls[[i]] <- dat.meanSF
    }
    sel.bkb.cent.lgc <- do.call(rbind, shifted.ls)
    saveRDS(sel.bkb.cent.lgc, file = file.path(paths["intermediary"], "sel.bkb.cent.lgc.rds"))
    }
    }
    message("\tCentred logicle backbone data... Obtained!")
    }
    
    message("\tDeriving initial clusters with PhenoGraph (forming the M matrix)...")
    a <- Sys.time()
    phenog.bkb.cent.lgc <- Rphenograph(
    sel.bkb.cent.lgc, 
    k = 50) 
    b <- Sys.time()
    message(b-a)
    saveRDS(phenog.bkb.cent.lgc, file = file.path(paths["intermediary"], "initM_phenog.bkb.cent.lgc.rds"))
    
    metadata.cell$init.M <- as.factor(membership(phenog.bkb.cent.lgc[[2]]))
    saveRDS(metadata.cell, file = file.path(paths["intermediary"], "fcs_metadata_df.rds"))
    
    if(plots==TRUE){
    
    message("\tUMAP with backbones (MPC)/proteins (FFC)...")
    a <- Sys.time()
    umap.bkb.cent.lgc <- umap(
    sel.bkb.cent.lgc, 
    n_neighbors = 15, min_dist = 0.2, metric = "euclidean", n_epochs = 2000)
    b <- Sys.time()
    message(b-a) 
    saveRDS(umap.bkb.cent.lgc, file = file.path(paths["intermediary"], "initM_umap.bkb.cent.lgc.rds"))
    
    
    message("\tVisualising clusters...")
    # pheno color
    col.v <- c(
    "#6E9A8A", "#6DE5DC", "#9237E9", "#B0E7BE", "#8761DB",
    "#71EA4A", "#9ACD49", "#E8813B", "#DEEB45", "#CED0EA",
    "salmon4", "#D8868B", "#EAE6C5", "green", "#E2AC7C",
    "#E050D7", "#887384", "#73AABE", "#DFD48C", "blue",
    "deeppink3", "mediumorchid4", "#ED574E", "#E5EB77", "#CEABDF",
    "#CAE7E6", "#6CAF76", "#5FE9B8", "#E0B341", "#B5E893",
    "#777FD5", "#64D4EB", "yellow")    ##33
    graph.dat <- data.frame(umap.bkb.cent.lgc, metadata.cell$init.M)
    colnames(graph.dat) <- c("UMAP1", "UMAP2", "Pheno.gp")
    
    n <- length(unique(graph.dat$Pheno.gp))
    qual_col_pals <- brewer.pal.info[brewer.pal.info$category == 'qual',,drop=FALSE]
    col.coeff <- unlist(mapply(brewer.pal, qual_col_pals$maxcolors, rownames(qual_col_pals)))[seq_len(n)]
    
    jpeg(filename = file.path(paths["graph"], paste0(trans.dat, "_UMAP_colPhenog_600x750.jpeg")), height = 600, width = 750, res = 80)
    p <- ggplot(
    data=graph.dat, mapping = aes(x=UMAP1, y=UMAP2, colour=Pheno.gp)) + geom_point(size = 0.05, alpha = 0.25) + 
    labs(titles = paste0("Initial biology (M matrix) from ", trans.dat," bkb. \n(", nrow(graph.dat)," cells)")) +
    guides(colour = guide_legend(override.aes = list(size=8))) + theme_bw() + 
    theme(text = element_text(size=22)) + scale_color_manual(values=col.coeff[seq_len(n)])
    plot(p)
    dev.off()
    }
    message("\tCompleted!")
    return(metadata.cell)
    }

#' Background correction for the backbone markers
#' 
#' This function has been designed to do background correction for the backbone markers by using normal-exponential convolution model.
#' 
#' @param paths a vector of characters of paths to store intput, intermediary results, outputs...
#' @param bkb.v a vector of the names of the backbone markers (MUST match to the names in the FCS file).
#' @param MPC if the data is from MPC experiments, set `MPC = TRUE`. Setting FALSE represents data from the fluorescence flow cytometry (FFC) assay.
#' @param bkb.upper.quantile the cut-off (default = 0.9) for selecting cells used for estimating the parameter of signal.
#' @param bkb.lower.quantile the cut-off (default = 0.1) for selecting cells used for estimating the parameters of noise.
#' @param bkb.min.quantile the cut-off (default = 0.01) for omitting the cells with the smallest values to minimise the impact of outliers.
#' @param plots logical; if TRUE (default), produce scatter plots for pre- and post- background adjusted backbone markers (calibrated values on y-axis and raw values on x-axis).
#' 
#' @author Hsiao-Chi Liao and Agus Salim
#' 
#' @importFrom grDevices dev.off jpeg
#' @importFrom graphics abline
#' @importFrom methods is
#' @importFrom stats as.formula contr.sum contrasts<- dexp dnorm median model.matrix optim pexp pnorm quantile sd setNames
#' @importFrom utils head read.csv
#' 
#' @return Background noise corrected backbone markers and graphs if specified
#' 
#' @details
#' Generating the calibrated measurements and save to medpara_bkc.bkb_no.bkcPhy_mt.rds file, and visualising the result with the scatter plots in the output directory.
#' 
bkc_bkb <-
function(
    paths,
    bkb.v,
    MPC,
    bkb.upper.quantile=0.9, #cells used for estimating parameter of signal
    bkb.lower.quantile=0.1, #cells used for estimating parameters of noise
    bkb.min.quantile=0.01, #the lowest 1% of values will not be used to minimise the impact of outliers on sig
    plots=TRUE
    ){
    ## for later use!
    upper.quantile <- bkb.upper.quantile
    lower.quantile <- bkb.lower.quantile
    min.quantile <- bkb.min.quantile
    
    rawInten <- readRDS(file = file.path(paths["intermediary"], "fcs_rawInten_mt.rds"))
    metadata.cell <- readRDS(file = file.path(paths["intermediary"], "fcs_metadata_df.rds"))
    sel.bkb.raw <- rawInten[,match(bkb.v, colnames(rawInten)),drop=FALSE]
    
    ## functions for estimating parameters
    {
    # nlogl for mu and sig parameters of noise distribution
    # xx is the observed intensity used for estimating mu and sig
    # the lowest 1% of values will not be used to minimize the impact of outliers on sig
    nlogl.norm.v2 <- function(p, xx, n, rm.min.quantile = min.quantile){
    xx <- sort(xx)[-seq_len(round(n*rm.min.quantile))]
    loglik <- sum(dnorm(xx,mean=p[1],sd=exp(p[2]),log=TRUE)) + 
    (n-length(xx))*pnorm(max(xx),mean=p[1],sd=exp(p[2]),log.p=TRUE,lower.tail=FALSE) +
    round(n*rm.min.quantile)*pnorm(min(xx),mean=p[1],sd=exp(p[2]),log.p=TRUE)
    -loglik
    }
    
    #nlogl for alpha parameter of the true signal
    # xx is the observed intensity used for estimating alpha
    nlogl.exp <- function(p, xx, n){
    loglik <- sum(dexp(xx,rate=exp(p[1]),log=TRUE)) + 
    (n-length(xx))*pexp(min(xx),rate=exp(p[1]),log.p=TRUE)
    -loglik
    }
    
    # input the obs intensity and return the calibrated intensity
    calib <- function(xx, alpha, mu, sig.2){
    mu.sx <- xx-mu-sig.2/alpha
    log.2ndterm <- dnorm(0,mean=mu.sx,sd=sqrt(sig.2),log=TRUE)-pnorm(0,mean=mu.sx,sd=sqrt(sig.2),lower.tail=FALSE,log.p=TRUE)
    calib.xx <- mu.sx + sig.2*exp(log.2ndterm)
    calib.xx
    }
    }
    
    ## estimating parameters (estimate phy's para but don't use them)
    {
    message("\tEstimating parameters for calibration...")
    para.ls <- list()
    for(m in seq_len(length(bkb.v))){
    if(MPC == TRUE){
    para.df <- data.frame(Unwanted = unique(metadata.cell$Well.lab), mu = NA, sig = NA, alpha = NA)
    }else{
    para.df <- data.frame(Unwanted = unique(metadata.cell$Batch), mu = NA, sig = NA, alpha = NA)
    }
    for(w in seq_len(length(para.df$Unwanted))){
    if(MPC == TRUE){
    dat.here <- sel.bkb.raw[which(metadata.cell$Well.lab == para.df$Unwanted[w]), m]
    }else{
    dat.here <- sel.bkb.raw[which(metadata.cell$Batch == para.df$Unwanted[w]), m]
    }
    ordering.bkb <- dat.here[order(dat.here)] #for finding quantiles
    
    ## estimate mean and sd of noise (10% smallest values)
    {
    first.quantile <- ordering.bkb[seq_len(length(ordering.bkb)*lower.quantile)]
    # initial values
    ini.mu <- mean(first.quantile)
    ini.log.sd <- log(sd(first.quantile))
    p.init <- c(ini.mu, ini.log.sd)
    
    suppressWarnings(
    out.1 <- tryCatch(optim(
    par=p.init,
    fn=nlogl.norm.v2,
    xx=first.quantile,
    n=length(ordering.bkb)),
    error = function(e){NA})
    )
    mle.mean <- out.1$par[1]
    mle.sd <- exp(out.1$par[2])
    
    para.df$mu[w] <- mle.mean
    para.df$sig[w] <- mle.sd
    }
    ## estimate log alpha using tryCatch, so it will not stop if some markers return error
    {
    last.quantile <- ordering.bkb[which(ordering.bkb > quantile(ordering.bkb, probs = upper.quantile))] #or +2*mle.sd
    p.init <- ifelse(mean(last.quantile) < 0, log(0.000001), log(mean(last.quantile)))
    suppressWarnings(
    out.2 <- tryCatch(optim(
    par=p.init,
    fn=nlogl.exp,
    xx=last.quantile,
    n=length(ordering.bkb),
    method = "Brent", lower = -20, upper = 0),
    error = function(e){NA})
    )

    alpha <- (1/exp(out.2$p))
    para.df$alpha[w] <- alpha
    }
    }
    para.ls[[m]] <- para.df
    message('backbone: ',m)
    }
    names(para.ls) <- colnames(sel.bkb.raw)
    if(MPC == TRUE){
    saveRDS(para.ls, file = file.path(paths["intermediary"], paste0("para.ls_",ncol(sel.bkb.raw),"bkb_",length(unique(metadata.cell$Well.lab)),"well.rds")))
    }else{
    saveRDS(para.ls, file = file.path(paths["intermediary"], paste0("para.ls_",ncol(sel.bkb.raw),"bkb_",length(unique(metadata.cell$Batch)),"batch.rds")))
    }
    message("\tEstimation of parameters... Completed!")
    }
    
    ## calibrating values
    {
    message("\tCalibrating backbone markers (except for physical measurements)...")
    calib.dat <- sel.bkb.raw
    {
    skip.phy.meas <- which(!(colnames(sel.bkb.raw) %in% c("FSC-A", "FSC-H", "FSC-W", "SSC-A", "SSC-H", "SSC-W")))
    for(m in skip.phy.meas){
    ## prepare 'median' parameters
    med.para <- apply(para.ls[[m]][,-1], 2, median)
    med.mle.mean <- med.para[1]
    med.mle.sd <- med.para[2]
    med.alpha <- med.para[3]
    
    calib.temp <- list()
    for(w in seq_len(length(para.df$Unwanted))){
    if(MPC == TRUE){
    dat.here <- sel.bkb.raw[which(metadata.cell$Well.lab == para.df$Unwanted[w]), m]
    }else{
    dat.here <- sel.bkb.raw[which(metadata.cell$Batch == para.df$Unwanted[w]), m]
    }
    ## calibrated values
    calib.temp[[w]] <- calib(dat.here, med.alpha, med.mle.mean, sig.2 = med.mle.sd^2)
    }
    #concatenating calib values for each "unwanted factor"
    calib.dat[,m] <- do.call(c, calib.temp)
    }
    saveRDS(calib.dat, file = file.path(paths["intermediary"], "medpara_bkc.bkb_no.bkcPhy_mt.rds"))
    
    ## scater plot: raw vs calib
    if(plots == TRUE){
    marker.nam <- colnames(calib.dat)
    for(i in seq_along(marker.nam)){
    jpeg(filename = file.path(paths["graph"], paste0("medpara_sub0.01.", marker.nam[i], "_para.", upper.quantile,"up.", lower.quantile,"lo_500x500.jpeg")), width = 500, height = 500, res = 80)
    #subsample for graphing
    extr.cell <- sample(seq_len(nrow(calib.dat)), nrow(calib.dat)*0.01)
    #can consider str. samp in the future
    if(i %in% skip.phy.meas){
    plot(
    x = sel.bkb.raw[extr.cell,i], y = calib.dat[extr.cell,i],
    xlab = "Raw", ylab = "Calib.", main = marker.nam[i], col = "blue",
    sub = paste0("Parameters Estimated from ", upper.quantile," (hi); ", lower.quantile," (lo)"))
    }else{
    plot(
    x = sel.bkb.raw[extr.cell,i], y = calib.dat[extr.cell,i],
    xlab = "Raw", ylab = "Raw.", main = marker.nam[i], col = "blue",
    sub = paste0("No need to calibrate the physical measurements."))
    }
    abline(a=0, b=1)
    dev.off()
    }
    }
    }
    message("\tCalibration of backbone markers... Completed!")
    }
    }

#' Background correction for the well-specific markers (PE)
#' 
#' This function has been designed to do background correction for the well-specific markers (PE) by using normal-exponential convolution model.
#' 
#' @param paths a vector of characters of paths to store intput, intermediary results, outputs...
#' @param pe.lower.quantile the cut-off (default = 0.1) for selecting cells used for estimating the parameters of noise for infinity markers.
#' @param pe.min.quantile the cut-off (default = 0.01) for omitting the cells with the smallest values to minimise the impact of outliers for infinity markers.
#' @param plots logical; if TRUE (default), produce scatter plots for pre- and post- background adjusted infinity markers (calibrated values on y-axis and raw values on x-axis).
#' 
#' @author Hsiao-Chi Liao and Agus Salim
#' 
#' @importFrom grDevices dev.off jpeg
#' @importFrom graphics abline
#' @importFrom methods is
#' @importFrom stats as.formula contr.sum contrasts<- dexp dnorm median model.matrix optim pexp pnorm quantile sd setNames
#' @importFrom utils head read.csv write.csv
#' 
#' @return Background noise corrected infinity markers and graphs if specified
#' 
#' @details
#' Generating the calibrated measurements and save to bkc.pe_mt.rds file, and visualising the result with the scatter plots in the output directory.
#' 
bkc_pe <-
function(
    paths,
    pe.lower.quantile=0.1, #cells used for estimating parameters of noise
    pe.min.quantile=0.01, #the lowest 1% of values will not be used to minimize the impact of outliers on sig
    plots=TRUE
    ){
    ## for later use!
    min.quantile <- pe.min.quantile
    lower.quantile <- pe.lower.quantile
    
    rawInten <- readRDS(file = file.path(paths["intermediary"], "fcs_rawInten_mt.rds"))
    metadata.cell <- readRDS(file = file.path(paths["intermediary"], "fcs_metadata_df.rds"))
    pe.raw <- as.matrix(rawInten[,"Legend"])
    colnames(pe.raw) <- "Legend"
    
    ## functions for estimating parameters
    {
    # nlogl for mu and sig parameters of noise distribution
    # xx is the observed intensity used for estimating mu and sig
    # the lowest 1% of values will not be used to minimize the impact of outliers on sig
    nlogl.norm.v2 <- function(p, xx, n, rm.min.quantile = min.quantile){
    xx <- sort(xx)[-seq_len(round(n*rm.min.quantile))]
    loglik <- sum(dnorm(xx,mean=p[1],sd=exp(p[2]),log=TRUE)) + 
    (n-length(xx))*pnorm(max(xx),mean=p[1],sd=exp(p[2]),log.p=TRUE,lower.tail=FALSE) +
    round(n*rm.min.quantile)*pnorm(min(xx),mean=p[1],sd=exp(p[2]),log.p=TRUE)
    -loglik
    }
    
    #nlogl for alpha parameter of the true signal
    # xx is the observed intensity used for estimating alpha
    nlogl.exp <- function(p, xx, n) {
    # exp(p[1]) = 1/exponential mean = 1/alpha
    loglik <- sum(dexp(xx,rate=exp(p[1]),log=TRUE)) + (n-length(xx))*pexp(min(xx),rate=exp(p[1]),log.p=TRUE)
    -loglik
    }
    
    # input the obs intensity and return the calibrated intensity
    calib <- function(xx, alpha, mu, sig.2) {
    mu.sx <- xx-mu-sig.2/alpha
    log.2ndterm <- dnorm(0,mean=mu.sx,sd=sqrt(sig.2),log=TRUE)-pnorm(0,mean=mu.sx,sd=sqrt(sig.2),lower.tail=FALSE,log.p=TRUE)
    calib.xx <- mu.sx + sig.2*exp(log.2ndterm)
    calib.xx
    }
    }
    
    ## estimating parameters and calibrating values
    message("\tEstimating parameters for calibration AND calibrating infinity markers...")
    para.df <- data.frame(Well.lab = unique(metadata.cell$Well.lab), mu = NA, sig = NA, alpha = NA)
    
    calib.dat.ls <- list()
    samp.alpha <- c()
    criteria.alpha <- c()
    few.cell.well.ls <- list()
    for(i in seq_along(para.df[,1])){
    well.here <- para.df$Well.lab[i]
    dat.here <- as.matrix(pe.raw[which(metadata.cell$Well.lab == well.here),1])
    
    #ordering according to size
    ordering.legend <- dat.here[,1][order(dat.here[,1])]
    
    ## estimate mean and var of noise (10% smallest values)
    first.quantile <- ordering.legend[seq_len(length(ordering.legend)*lower.quantile)]
    # initial values
    ini.mu <- mean(first.quantile)
    ini.log.sd <- log(sd(first.quantile))
    p.init <- c(ini.mu, ini.log.sd)
    
    suppressWarnings(
    out.1 <- tryCatch(optim(
    par=p.init,
    fn=nlogl.norm.v2,
    xx=first.quantile,
    n=length(ordering.legend)),
    error = function(e){NA})
    )
    
    mle.mean <- out.1$par[1]
    mle.sd <- exp(out.1$par[2])
    
    para.df$mu[i] <- mle.mean
    para.df$sig[i] <- mle.sd
    
    ## estimate log alpha using tryCatch, so it will not stop if some markers return error
    ## orig with large no cell: mle.mean+3*mle.sd -> mle.mean+1*mle.sd
    last.mu.3sd <- ordering.legend[which(ordering.legend > (mle.mean+3*mle.sd))]
    if(length(last.mu.3sd) < 10){
    last.mu.3sd <- ordering.legend[which(ordering.legend >= head(ordering.legend,10))]
    criteria.alpha[i] <- "largest10"
    few.cell.well.ls[[i]] <- well.here
    }else{
    criteria.alpha[i] <- "mean+3sd"
    }
    samp.alpha[i] <- length(last.mu.3sd)
    p.init <- ifelse(mean(last.mu.3sd) < 0, log(0.000001), log(mean(last.mu.3sd)))
    suppressWarnings(
    out.2 <- tryCatch(optim(
    par=p.init,
    fn=nlogl.exp,
    xx=last.mu.3sd,
    n=length(ordering.legend),
    method = "Brent", lower = -20, upper = 0),
    error = function(e){NA})
    )
    
    alpha <- (1/exp(out.2$p))
    para.df$alpha[i] <- alpha
    
    ## calibrating values
    calib.dat <- dat.here
    calib.dat[,1] <- calib(dat.here, alpha, mle.mean, sig.2 = mle.sd^2)
    calib.dat.ls[[i]] <- calib.dat
    
    }
    message("Could not find enough cells (>=10) when used \"mle.mean+3*mle.sd\", so estimated alpha with the top 10 cells with \"the largest values\":\n", length(do.call(c, few.cell.well.ls)), " wells applied this strategy\nSee Wellname_largest10.csv in the intermediary directory for details.")
    out_largest10 <- data.frame(do.call(c, few.cell.well.ls))
    colnames(out_largest10) <- "Well"
    write.csv(out_largest10, file = file.path(paths["intermediary"], "Wellname_largest10.csv"))
    
    para.df <- cbind(para.df, samp.alpha, criteria.alpha)
    saveRDS(para.df, file = file.path(paths["intermediary"], "para.266pe.rds"))
    
    calib.dat <- do.call(rbind, calib.dat.ls); colnames(calib.dat) <- "Legend"
    
    saveRDS(calib.dat, file = file.path(paths["intermediary"], "bkc.pe_mt.rds"))
    
    ## scater plot: raw vs calib
    if(plots==TRUE){
    for(i in seq_along(para.df$Well.lab)){
    jpeg(filename = file.path(paths["graph"], paste0("pe_", para.df$Well.lab[i], "_para.", para.df$criteria.alpha[i], "up.", lower.quantile,"lo_500x500.jpeg")), width = 500, height = 500, res = 80)
    #can consider str. samp in the future
    extr.cell <- which(metadata.cell$Well.lab == para.df$Well.lab[i])
    plot(
    x = pe.raw[extr.cell,1], y = calib.dat[extr.cell,1],
    xlab = "Raw", ylab = "Calib.", main = para.df$Well.lab[i], col = "darkgoldenrod4", 
    sub = paste0("Parameters Estimated from ", criteria.alpha[i], " (", samp.alpha[i]," cells (hi)); ", lower.quantile," (lo)"))
    abline(a=0, b=1)
    dev.off()
    }
    }
    message("\tCalibration of infinity markers... Completed!")
    
    return(calib.dat)
    }

#' Making the input for imputation
#' 
#' This function has been designed to combine the normalised backbone measurements and the normalised PE markers for later imputation.
#' 
#' @param paths a vector of characters of paths to store intput, intermediary results, outputs...
#' 
#' @author Hsiao-Chi Liao
#' 
#' @importFrom grDevices dev.off jpeg
#' @importFrom graphics abline
#' @importFrom methods is
#' @importFrom stats as.formula contr.sum contrasts<- dexp dnorm median model.matrix optim pexp pnorm quantile sd setNames
#' @importFrom utils head read.csv
#' 
#' @return Combined normalised backbone and infinity markers for imputation
#' 
#' @details
#' Generating the combined data and saving to impu.input_log.mt.rds (on log scale) in the output directory.
#' 
mkImputeMT <-
function(
    paths
    ){
    bkc.adj.bkb <- readRDS(file = file.path(paths["downstream"], "bkc.adj.bkb_logScale_mt.rds")) #no. of selected bkb
    bkc.pe <- readRDS(file = file.path(paths["intermediary"], "bkc.pe_mt.rds"))
    metadata.cell <- readRDS(file = file.path(paths["intermediary"], "fcs_metadata_df.rds"))
    
    impu.input_log.mt <- cbind(bkc.adj.bkb, log(bkc.pe))
    rownames(impu.input_log.mt) <- metadata.cell$NO.in.all
    
    saveRDS(impu.input_log.mt, file = file.path(paths["intermediary"], "impu.input_log.mt.rds"))
    }
    

#' Converting FCS files to RDS files
#'
#' This function has been designed to convert the raw FCS files to data matrix and export to RDS files.
#'
#' @param paths a vector of characters of paths to store intput, intermediary results, outputs...
#' @param file_meta if the file names of the FCS files are in the specified format, set file_meta="auto"; otherwise set file_meta="usr" and provide "filename_meta.csv" in FCSpath.
#' @param yvar the name of the well-specific marker in the FCS files (e.g., "Legend").
#'
#' @author Hsiao-Chi Liao
#'
#' @importFrom flowCore read.flowSet keyword parameters
#' @importFrom Biobase pData exprs
#' @importFrom stringr str_extract
#' 
#' @importFrom grDevices dev.off jpeg
#' @importFrom graphics abline
#' @importFrom methods is
#' @importFrom stats as.formula contr.sum contrasts<- dexp dnorm median model.matrix optim pexp pnorm quantile sd setNames
#' @importFrom utils head read.csv
#'
#' @return Raw protein intensities and the corresponding metadata from MPC experiments
#'
#' @details
#' Generating fcs_metadata_df.rds and fcs_rawInten_mt.rds files in the output directory.
#'
fcs_to_rds <-
function(
    paths, file_meta, yvar
    ){
    ## re-make large mt and consider SSC and FSC
    # import fcs.impu.raw files ###
    fs <- read.flowSet(path = file.path(paths["input"], "fcs"))
    name.desc.fs <- pData(parameters(fs[[1]]))
    name.desc.fs$desc[which(is.na(name.desc.fs$desc))] <- name.desc.fs$name[which(is.na(name.desc.fs$desc))]
    
    ####
    
    #### data - raw ####
    fcs.raw.mt.list <- fcs.raw.meta.list <- list()
    if(file_meta == "auto"){
    for(i in seq_along(fs)){
    fcs.raw.mt.list[[i]] <- exprs(fs[[i]])
    fcs.raw.meta.list[[i]] <- cbind(
    NO.in.well=seq_len(dim(fcs.raw.mt.list[[i]])[1]),
    Filenam=unlist(keyword(fs[[i]], "GUID")))
    }
    fcs.raw.mt <- do.call(rbind, fcs.raw.mt.list)
    colnames(fcs.raw.mt) <- name.desc.fs$desc[match(colnames(fcs.raw.mt), name.desc.fs$name)]
    ####
    
    # making metadata ###
    fcs.raw.meta.df <- as.data.frame(do.call(rbind, fcs.raw.meta.list)) 
    fcs.raw.meta.df$NO.in.well <- as.integer(fcs.raw.meta.df$NO.in.well)
    #use lapply for multiple changing class: lapply(mydf[,2:3], as.factor)
    
    #plate
    fcs.raw.meta.df$Plate <- str_extract(fcs.raw.meta.df$Filenam, "Plate[0-9]")
    
    #well
    fcs.raw.meta.df$Well <- str_extract(fcs.raw.meta.df$Filenam, "[A-H]\\d{1,2}") #[:digit:], or the shorthand \\d
    orig.lab <- c(
    paste0(rep(c("A","B","C","D","E","F","G","H"), each = 9), seq_len(9)),
    paste0(rep(c("A","B","C","D","E","F","G","H"), each = 9), c(10:12)))
    
    new.lab <- c(
    paste0(rep(c("A","B","C","D","E","F","G","H"), each = 9), 0, seq_len(9)),
    paste0(rep(c("A","B","C","D","E","F","G","H"), each = 9), c(10:12)))
    map <- setNames(new.lab, orig.lab)
    fcs.raw.meta.df$Well <- map[fcs.raw.meta.df$Well]
    
    #column
    fcs.raw.meta.df$Column <- paste0("Col.", sub("[A-H]", "", fcs.raw.meta.df$Well))
    
    #row
    fcs.raw.meta.df$Row <- sub("\\d{1,2}", "", fcs.raw.meta.df$Well)
    orig.lab <- c("A","B","C","D","E","F","G","H")
    new.lab <- paste0("Row.", c("01","02","03","04","05","06","07","08"))
    map <- setNames(new.lab, orig.lab)
    fcs.raw.meta.df$Row <- map[fcs.raw.meta.df$Row]
    
    #well.lab
    fcs.raw.meta.df$Well.lab <- paste0(sub("late", "", fcs.raw.meta.df$Plate),"_",fcs.raw.meta.df$Well)
    
    if(sum(is.na(fcs.raw.meta.df)) != 0){
    message("\tfile_meta='auto' doesn't work on your data.")
    stop("\tPlease provide a \"filename_meta.csv\" file in `FCSpath/meta/` and use `file_meta=\"usr\"` instead.")
    }
    
    #ordering before giving NO.in.all
    ord.fcs.raw.meta.df <- fcs.raw.meta.df[order(fcs.raw.meta.df$Well.lab),,drop=FALSE]
    ord.fcs.raw.mt <- fcs.raw.mt[order(fcs.raw.meta.df$Well.lab),,drop=FALSE]
    }
    
    if(file_meta == "usr"){
    meta.inf <- read.csv(file = file.path(paths["input"], "meta", "filename_meta.csv"))
    for(i in seq_along(fs)){
    fcs.raw.mt.list[[i]] <- exprs(fs[[i]]) #22 columns
    filenam <- unlist(keyword(fs[[i]], "GUID"))
    fcs.raw.meta.list[[i]] <- cbind(
    NO.in.well=(seq_len(dim(fcs.raw.mt.list[[i]])[1])),
    Filenam=filenam,
    Plate=meta.inf[which(filenam == meta.inf$Filenam), "Plate"],
    Well=meta.inf[which(filenam == meta.inf$Filenam), "Well"],
    Column=meta.inf[which(filenam == meta.inf$Filenam), "Column"],
    Row=meta.inf[which(filenam == meta.inf$Filenam), "Row"],
    Well.lab=meta.inf[which(filenam == meta.inf$Filenam), "Well.lab"])
    }
    fcs.raw.mt <- do.call(rbind, fcs.raw.mt.list) 
    colnames(fcs.raw.mt) <- name.desc.fs$desc[match(colnames(fcs.raw.mt), name.desc.fs$name)]
    ####
    
    # making metadata ###
    fcs.raw.meta.df <- as.data.frame(do.call(rbind, fcs.raw.meta.list)) 
    fcs.raw.meta.df$NO.in.well <- as.integer(fcs.raw.meta.df$NO.in.well)
    #use lapply for multiple changing class: lapply(mydf[,2:3], as.factor)
    
    #ordering before giving NO.in.all
    ord.fcs.raw.meta.df <- fcs.raw.meta.df[order(fcs.raw.meta.df$Well.lab),,drop=FALSE]
    ord.fcs.raw.mt <- fcs.raw.mt[order(fcs.raw.meta.df$Well.lab),,drop=FALSE]
    }
    
    #global ID for each cell
    ord.fcs.raw.meta.df$NO.in.all <- seq_len(nrow(ord.fcs.raw.mt))
    
    # name exploratory as Legend
    colnames(ord.fcs.raw.mt)[which(colnames(ord.fcs.raw.mt) == yvar)] <- "Legend"
    
    #out rds
    ord.fcs.raw.meta.df.out <- ord.fcs.raw.meta.df[,c(8,1,3,5,6,4,7),drop=FALSE]
    saveRDS(ord.fcs.raw.meta.df.out, file = file.path(paths["intermediary"], "fcs_metadata_df.rds"))
    saveRDS(ord.fcs.raw.mt, file = file.path(paths["intermediary"], "fcs_rawInten_mt.rds"))
    }

#' Normalising data from the Fluorescence Flow Cytometry (FFC) Experiments with mapfx.norm
#'
#' @description This function is used to normalise, including background correction and removal of batch effects, protein intensity data from FFC assays. The input data is in FCS format. The functions include data normalisation and cluster analysis.
#'
#' @param runVignette logical; if FALSE (default), specify a path to `FCSpath` argument; TRUE for running Vignette using built-in data.
#' @param runVignette_meta the argument for the built-in metadata when running Vignette; NULL (default).
#' @param runVignette_rawInten the argument for the built-in raw intensities when running Vignette; NULL (default).
#' @param FCSpath path to the input directory where `filename_meta.csv` and FCS files are stored. `filename_meta.csv` should be saved under `FCSpath/FCS/meta/` and FCS files should be saved under `FCSpath/FCS/fcs/`. See Vignette for details.
#' @param Outpath path to the output directory where intermediate results and final results will be stored.
#' @param protein.v a vector of the names of the protein markers (MUST be the same as the names in the FCS files). For example, `protein.v = c("FSC-H","FSC-W","SSC-H","SSC-W","CD3","CD4","CD8","CD45")`.
#' @param protein.upper.quantile the cut-off (default = 0.9) for selecting cells used for estimating the parameter of signal for protein markers.
#' @param protein.lower.quantile the cut-off (default = 0.1) for selecting cells used for estimating the parameters of noise for protein markers.
#' @param protein.min.quantile the cut-off (default = 0.01) for omitting the cells with the smallest values to minimise the impact of outliers during estimation.
#' @param plots.bkc.protein logical; if TRUE (default), produce scatter plots for pre- and post- background adjusted protein markers (calibrated values on y-axis and raw values on x-axis).
#' @param plots.initM logical; if TRUE (default), produce an UMAP embedding plot to visualise the structure of the biological clusters used to form the initial M matrix for removal of batch effects.
#' @param plots.rmBatchEffect logical; if TRUE (default), produce heatmaps to visualise the unwanted (batch) effects and biological effects in the pre- and post- adjusted datasets.
#' @param cluster.analysis.protein  logical; if TRUE (default), perform cluster analysis using normalised protein markers.
#' @param plots.cluster.analysis.protein logical; if TRUE (default), produce an UMAP embedding plot from the normalised protein markers to visualise the structure of the biological clusters.
#'
#' @author Hsiao-Chi Liao, Agus Salim
#'
#' @importFrom utils data
#'
#' @export
#' 
#' @return Normalised protein markers on log scale and metadata for cells
#' 
#' @details
#' In the output directory, this function produces the normalised protein measurements, cell group labels from the cluster analysis using normalised proteins, and graphs will be provided if specified.
#' 
#' @examples
#' # import built-in data
#' data(ord.fcs.raw.meta.df.out_ffc)
#' data(ord.fcs.raw.mt_ffc)
#' 
#' # create an Output directory for the MapfxFFC function
#' dir.create(file.path(tempdir(), "FFCnorm_Output"))
#' 
#' MapfxFFC_obj <- MapfxFFC(
#'   runVignette = TRUE, #set FALSE if not running this example
#'   runVignette_meta = ord.fcs.raw.meta.df.out_ffc, #set NULL if not running this example
#'   runVignette_rawInten = ord.fcs.raw.mt_ffc, #set NULL if not running this example
#'   FCSpath = NULL, # users specify their own input path if not running this example
#'   Outpath = file.path(tempdir(), "FFCnorm_Output"),
#'   protein.v = c("CD3","CD4","CD8","CD45"),
#'   protein.upper.quantile = 0.9, 
#'   protein.lower.quantile = 0.1, 
#'   protein.min.quantile = 0.01,
#'   plots.bkc.protein = TRUE,
#'   plots.initM = TRUE,
#'   plots.rmBatchEffect = TRUE,
#'   cluster.analysis.protein = TRUE, plots.cluster.analysis.protein = TRUE)
#'   
#'
MapfxFFC <-
function(
    runVignette = FALSE,
    runVignette_meta = NULL,
    runVignette_rawInten = NULL,
    FCSpath = NULL,
    Outpath = NULL,
    protein.v = NULL,
    protein.upper.quantile = 0.9, 
    protein.lower.quantile = 0.1, 
    protein.min.quantile = 0.01,
    plots.bkc.protein = TRUE,
    plots.initM = TRUE,
    plots.rmBatchEffect = TRUE,
    cluster.analysis.protein = TRUE, plots.cluster.analysis.protein = TRUE){
    
    #run Vignette
    if(runVignette == TRUE){
    if(is.null(Outpath)){
    message("\nPlease make sure the output path has been specified!")
    }
    
    message("\n\n\nCreating directories for output...")
    settings <- initialize(
    path_to_fcs=FCSpath,
    path_to_output=Outpath,
    verbose=TRUE)
    paths <- settings$paths
    
    ###
    
    saveRDS(runVignette_meta, file = file.path(paths["intermediary"], "fcs_metadata_df.rds"))
    saveRDS(runVignette_rawInten, file = file.path(paths["intermediary"], "fcs_rawInten_mt.rds"))
    }
    
    #not running Vignette
    if(runVignette == FALSE){
    
    if(is.null(FCSpath) | is.null(Outpath) | is.null(protein.v)){
    stop("\nPlease make sure you have specified\n(1) path to the input files (details in the folder diagram in the vignette) \n(2) output path \n(3) a list of protein markers")
    }
    
    if(!is.null(FCSpath) & !is.null(Outpath) & !is.null(protein.v)){
    
    message("\n\n\nCreating directories for output...")
    
    settings <- initialize(
    path_to_fcs=FCSpath,
    path_to_output=Outpath,
    verbose=TRUE)
    paths <- settings$paths
    
    ###
    
    ##1.fcs to rds
    message("\n\n\nTransforming FCS to RDS files...")
    #below function for reading ffc data only
    fcs_to_rds_bkb(
    paths=paths, file_meta="usr", MPC = FALSE)
    }
    }
    
    ##2-1.bkc protein
    message("\n\n\nBackground correcting proteins...")
    bkc_bkb(
    paths=paths, bkb.v=protein.v,
    MPC=FALSE,
    bkb.upper.quantile=protein.upper.quantile, #cells used for estimating parameter of signal
    bkb.lower.quantile=protein.lower.quantile, #cells used for estimating parameters of noise
    bkb.min.quantile=protein.min.quantile,
    plots=plots.bkc.protein) #the lowest 1% of values will not be used to minimise the impact of outliers on sig
    
    ##3-0.initM
    message("\n\n\nForming a matrix of biology (M) for removal of batch effect...")
    metadata <- initM(
    paths=paths, assay="FFC", bkb.v=protein.v,
    plots=plots.initM) #may give "cent.log.bkc" in the next version
    
    ##3-1.rmBatchEffect
    message("\n\n\nRemoval of batch effect...")
    norm.bkb <- rmBatchEffect(
    paths=paths,
    plots = plots.rmBatchEffect)
    
    ##7.cluster analysis proteinOnly
    if(cluster.analysis.protein == TRUE){
    message("\n\n\nCluster analysis with adjusted protein markers for ALL cells...")
    metadata <- cluster.analysis.bkbOnly(
    paths=paths,
    bkb.v=protein.v,
    plots = plots.cluster.analysis.protein)
    }
    message("\nCell group labels are saved in \"GP.denoised.bkb.allCells\" column...")
    
    message("\tCompleted!")
    return(list(logNormalisedMarkers=norm.bkb, Metadata=metadata))
    }

# Checks MapfxFFC
data(ord.fcs.raw.meta.df.out_ffc)
data(ord.fcs.raw.mt_ffc)

# create an Output directory in the current working directory for the argument 'Outpath' of the MapfxFFC function
dir.create(file.path(tempdir(), "FFCnorm_Output"))

MapfxFFC_obj <- MapfxFFC(
  runVignette = TRUE, #set FALSE if not running this example
  runVignette_meta = ord.fcs.raw.meta.df.out_ffc, #set NULL if not running this example
  runVignette_rawInten = ord.fcs.raw.mt_ffc, #set NULL if not running this example
  FCSpath = NULL, # users specify their own input path if not running this example
  Outpath = file.path(tempdir(), "FFCnorm_Output"),
  protein.v = c("CD3","CD4","CD8","CD45"),
  protein.upper.quantile = 0.9,
  protein.lower.quantile = 0.1,
  protein.min.quantile = 0.01,
  plots.bkc.protein = TRUE,
  plots.initM = TRUE,
  plots.rmBatchEffect = TRUE,
  cluster.analysis.protein = TRUE, plots.cluster.analysis.protein = TRUE)

test_that("MapfxFFC returns an object", {
  obj <- MapfxFFC_obj
  expect_true(validObject(obj))
})

test_that("MapfxFFC returns a list", {
  obj <- MapfxFFC_obj
  expect_true(inherits(obj, "list"))
})

test_that("MapfxFFC returns a list with size 2", {
  obj <- MapfxFFC_obj
  expect_length(obj, 2)
})

#' Removing well effect from the data
#' 
#' This function has been designed to remove the unwanted effects (well effects) from the background corrected measurements.
#' 
#' @param paths a vector of characters of paths to store intput, intermediary results, outputs...
#' @param plots logical; if TRUE (default), produce heatmaps to visualise the unwanted (well) effects and biological effects in the pre- and post- adjusted datasets.
#' 
#' @author Hsiao-Chi Liao
#' 
#' @import ggplot2
#' @importFrom ComplexHeatmap rowAnnotation draw Heatmap
#' @importFrom Rfast lmfit colsums
#' @importFrom RColorBrewer brewer.pal.info brewer.pal
#' @importFrom circlize colorRamp2
#' 
#' @importFrom grDevices dev.off jpeg
#' @importFrom graphics abline
#' @importFrom methods is
#' @importFrom stats as.formula contr.sum contrasts<- dexp dnorm median model.matrix optim pexp pnorm quantile sd setNames
#' @importFrom utils head read.csv
#' 
#' @return Normalised backbone markers on log scale
#' 
#' @details
#' Generating the calibrated measurements and saving to bkc.adj.bkb_logScale_mt.rds (on log scale) and bkc.adj.bkb_linearScale_mt.rds (on linear scale), and visualising the result with the heatmaps in the output directory.
#' 
rmWellEffect <-
function(
    paths, 
    plots=TRUE
    ){
    bkc.bkb <- readRDS(file = file.path(paths["intermediary"], "medpara_bkc.bkb_no.bkcPhy_mt.rds")) #no. of selected bkb
    metadata.cell <- readRDS(file = file.path(paths["intermediary"], "fcs_metadata_df.rds"))
    
    ### Rfast to estimate coef for later adj. ###
    {
    message("\tEstimating coefficients for removing well effect (Rfast - pre.adj)...")
    {
    metadata.cell[,c("Plate","Column","Row","init.M")] <- lapply(metadata.cell[,c("Plate","Column","Row","init.M")], factor)
    contrasts(metadata.cell[,"Plate"]) <- contr.sum(3, contrasts=TRUE)
    contrasts(metadata.cell[,"Column"]) <- contr.sum(12, contrasts=TRUE)
    contrasts(metadata.cell[,"Row"]) <- contr.sum(8, contrasts=TRUE)
    contrasts(metadata.cell[,"init.M"]) <- contr.sum(length(unique(metadata.cell$init.M)), contrasts=TRUE)
    
    xx <- as.matrix(model.matrix(~ Plate + Column + Row + init.M, data = metadata.cell))
    
    lm.results.ls <- list()
    ln.sig.v <- c()
    res.df <- bkc.bkb
    for(i in seq_along(bkc.bkb[1,])){
    yy.val <- as.matrix(bkc.bkb[,i])
    
    lm.result <- tryCatch(lmfit(x = xx, y = log(yy.val+0.000001)), error=function(err) NA)
    lm.results.ls[[i]] <- lm.result
    res <- lm.result$residuals
    ln.sig.v[i] <- sqrt(sum( (res - mean(res) )^2 )/(length(res)-length(lm.results.ls[[1]]$be))) #coef used

    res.df[,i] <- res
    
    message("Processing backbone: ", i)
    }
    
    names(lm.results.ls) <- colnames(bkc.bkb)
    
    quant.bio.pcr.mt <- do.call(cbind, lapply(lm.results.ls, '[[', 1))
    colnames(quant.bio.pcr.mt) <- names(lm.results.ls)
    quant.bio.pcr.mt <- rbind(quant.bio.pcr.mt, ln.sig = ln.sig.v)
    save(lm.results.ls, quant.bio.pcr.mt, file = file.path(paths["intermediary"], "rmWellEffect_log.normal.reg_coef.mdl_pre.RData"))
    save(res.df, file = file.path(paths["intermediary"], "rmWellEffect_log.normal.reg_residuals.RData"))
    }
    message("\tEstimation completed!")
    }
    ### adjusting data - subtracting unwanted effect ###
    {
    message("\tRemoving well effect for backbone markers...")
    adj.data <- log.adj.data <- bkc.bkb
    for(i in seq_along(bkc.bkb[1,])){
    unwanted.eff <- (xx[,(2:21)] %*% quant.bio.pcr.mt[(2:21), i])
    
    log.adj.data[,i] <- (log(bkc.bkb[,i]+0.000001) - unwanted.eff) #as the coef was estimated from data on log scale
    
    adj.data[,i] <- (exp(log.adj.data[,i])-0.000001)
    }
    saveRDS(log.adj.data, file = file.path(paths["downstream"], "bkc.adj.bkb_logScale_mt.rds"))
    saveRDS(adj.data, file = file.path(paths["downstream"], "bkc.adj.bkb_linearScale_mt.rds"))
    message("\tAdjustment completed!")
    }
    ### Rfast to estimate coef for later adj. ###
    {
    message("\tExamining the existence of well effect in the adjusted data (Rfast - post.adj)...")
    {
    metadata.cell[,c("Plate","Column","Row","init.M")] <- lapply(metadata.cell[,c("Plate","Column","Row","init.M")], factor)
    contrasts(metadata.cell[,"Plate"]) <- contr.sum(3, contrasts=TRUE)
    contrasts(metadata.cell[,"Column"]) <- contr.sum(12, contrasts=TRUE)
    contrasts(metadata.cell[,"Row"]) <- contr.sum(8, contrasts=TRUE)
    contrasts(metadata.cell[,"init.M"]) <- contr.sum(length(unique(metadata.cell$init.M)), contrasts=TRUE)
    
    xx <- as.matrix(model.matrix(~ Plate + Column + Row + init.M, data = metadata.cell))
    
    adj.lm.results.ls <- list()
    ln.sig.v <- c()
    for(i in seq_along(log.adj.data[1,])){
    yy.val <- as.matrix(log.adj.data[,i])
    
    lm.result <- tryCatch(lmfit(x = xx, y = yy.val), error=function(err) NA)
    adj.lm.results.ls[[i]] <- lm.result
    res <- lm.result$residuals
    ln.sig.v[i] <- sqrt(sum( (res - mean(res) )^2 )/(length(res)-length(lm.results.ls[[1]]$be))) #coef used

    message("Processing backbone: ", i)
    }
    
    names(adj.lm.results.ls) <- colnames(log.adj.data)
    
    adj.quant.bio.pcr.mt <- do.call(cbind, lapply(adj.lm.results.ls, '[[', 1))
    colnames(adj.quant.bio.pcr.mt) <- names(adj.lm.results.ls)
    adj.quant.bio.pcr.mt <- rbind(adj.quant.bio.pcr.mt, ln.sig = ln.sig.v)
    save(lm.results.ls, quant.bio.pcr.mt, adj.quant.bio.pcr.mt, file = file.path(paths["intermediary"], "rmWellEffect_log.normal.reg_coef.mdl_post.RData"))
    }
    }
    
    ### rmWellEffect_bkb_visualisation (heatmap) - bio and pcr effects ###
    if(plots == TRUE){
    {
    #adjust y the num. one
    #no intercept and sigma
    in.dat <- list(quant.bio.pcr.mt[-c(1,nrow(quant.bio.pcr.mt)),,drop=FALSE], adj.quant.bio.pcr.mt[-c(1,nrow(adj.quant.bio.pcr.mt)),,drop=FALSE])
    in.dat <- list(
    rbind(
    in.dat[[1]][seq_len(2),,drop=FALSE], Plate3 = -colsums(in.dat[[1]][seq_len(2),,drop=FALSE]),
    in.dat[[1]][3:13,,drop=FALSE], Column12 = -colsums(in.dat[[1]][3:13,,drop=FALSE]),
    in.dat[[1]][14:20,,drop=FALSE], Row8 = -colsums(in.dat[[1]][14:20,,drop=FALSE]),
    in.dat[[1]][21:nrow(in.dat[[1]]),,drop=FALSE], init.Mlast = -colsums(in.dat[[1]][21:nrow(in.dat[[1]]),,drop=FALSE])),
    rbind(
    in.dat[[2]][seq_len(2),,drop=FALSE], Plate3 = -colsums(in.dat[[2]][seq_len(2),,drop=FALSE]),
    in.dat[[2]][3:13,,drop=FALSE], Column12 = -colsums(in.dat[[2]][3:13,,drop=FALSE]),
    in.dat[[2]][14:20,,drop=FALSE], Row8 = -colsums(in.dat[[2]][14:20,,drop=FALSE]),
    in.dat[[2]][21:nrow(in.dat[[2]]),,drop=FALSE], init.Mlast = -colsums(in.dat[[2]][21:nrow(in.dat[[2]]),,drop=FALSE]))
    )
    rownames(in.dat[[1]])[length(rownames(in.dat[[1]]))] <- paste0("init.M", (length(24:nrow(in.dat[[1]]))))
    rownames(in.dat[[2]])[length(rownames(in.dat[[2]]))] <- paste0("init.M", (length(24:nrow(in.dat[[2]]))))
    
    ## consistent colour key
    bio.lim <- c(min(in.dat[[1]][-seq_len(23),], in.dat[[2]][-seq_len(23),]), max(in.dat[[1]][-seq_len(23),], in.dat[[2]][-seq_len(23),]))
    well.lim <- c(min(in.dat[[1]][seq_len(23),], in.dat[[2]][seq_len(23),]), max(in.dat[[1]][seq_len(23),], in.dat[[2]][seq_len(23),]))
    
    pre.post <- c("pre-", "post-")
    bio.sys <- c("bio.", "well.")
    for(i in seq_along(pre.post)){ #pre & post
    for(j in seq_along(bio.sys)){ #bio & batch
    ### heatmap
    dat0 <- in.dat[[i]]
    
    if(j == 1){ #biology
    bio.dat <- dat0[-seq_len(23),,drop=FALSE]
    # metadata
    meta_mat <- data.frame(coeff=rownames(bio.dat))
    meta_mat$Coef <- factor(
    paste0('init.M', seq_len(nrow(meta_mat))), 
    levels = paste0('init.M', seq_len(nrow(meta_mat))))
    rownames(meta_mat) <- meta_mat$Coef
    
    n <- nrow(meta_mat)
    qual_col_pals <- brewer.pal.info[brewer.pal.info$category == 'qual',]
    col.coeff <- unlist(
    mapply(brewer.pal, qual_col_pals$maxcolors, rownames(qual_col_pals)))[seq_len(n)]
    names(col.coeff) <- meta_mat$Coef
    row_ha <- rowAnnotation(
    Effect=meta_mat[,'Coef'],
    col=list(Effect=col.coeff), show_legend = FALSE)
    #elements in `col` should be named vectors.
    
    # color
    coeff_col_func <- colorRamp2(c(bio.lim[1], 0, bio.lim[2]), c("blue", "gray0", "darkorange1"))
    
    #not cluster rows/ cols
    jpeg(file.path(paths["graph"], paste0("Heatmap_coeffs_", pre.post[i], bio.sys[j], "_2000x1500.jpeg")), height = 2000, width = 1500, res = 250)
    draw(Heatmap(bio.dat, name=paste0("MLE"), #Rfast
    column_title = paste0(
    pre.post[i], "well.adj."),
    row_title = bio.sys[j],
    right_annotation = row_ha,
    col = coeff_col_func,
    column_dend_height = unit(1,"cm"),  # tree size
    cluster_rows = FALSE,
    cluster_columns = FALSE,
    clustering_method_rows = "average",
    clustering_method_columns = "average",
    show_column_names = TRUE,
    show_row_names = TRUE)
    )
    dev.off()
    }else{
    well.dat <- dat0[(seq_len(23)),,drop=FALSE]
    # metadata
    meta_mat <- data.frame(coeff=rownames(well.dat))
    meta_mat$Coef <- factor(c(
    rep('logn.mu.Plate', 3), rep('logn.mu.Column', 12), rep('logn.mu.Row', 8)), 
    levels = c('logn.mu.Plate', 'logn.mu.Column', 'logn.mu.Row'))
    
    # col. anno for heatmap (be careful of the position of each variable)
    col.coeff <- c(brewer.pal(9,"Set1")[c(1,3,2)], brewer.pal(8,"Dark2")[c(4,1,3)])
    row_ha <- rowAnnotation(
    Effect=meta_mat[,'Coef'],
    col=list(Effect=c(
    'logn.mu.Plate' = col.coeff[4],
    'logn.mu.Column' = col.coeff[5],
    'logn.mu.Row' = col.coeff[6])))
    # color
    coeff_col_func <- colorRamp2(c(well.lim[1], 0, well.lim[2]), c("blue", "gray0", "darkorange1"))
    
    #not cluster rows/ cols
    jpeg(file.path(paths["graph"], paste0("Heatmap_coeffs_", pre.post[i], bio.sys[j], "_2000x1500.jpeg")), height = 2000, width = 1500, res = 250)
    draw(Heatmap(
    well.dat, name=paste0("MLE"), #Rfast
    column_title = paste0(pre.post[i], "well.adj."),
    row_title = bio.sys[j],
    right_annotation = row_ha,
    col = coeff_col_func,
    column_dend_height = unit(1,"cm"),  # tree size
    cluster_rows = FALSE,
    cluster_columns = FALSE,
    clustering_method_rows = "average",
    clustering_method_columns = "average",
    show_column_names = TRUE,
    show_row_names = TRUE)
    )
    dev.off()
    }
    }
    }
    }
    }
    
    return(log.adj.data)
    }

library(flowCore)
## Using the Example Datasets in MAPFX Package for the Vignette
### MPC
# This dataset is a subset of the single-cell murine lung data at steady state downloaded from 
# [FlowRepository](https://flowrepository.org/id/FR-FCM-Z2LP) provided by Etienne Becht (Nov 2020). 
# The raw protein intensities and the corresponding metadata were saved in 
# `data/ord.fcs.raw.mt_mpc.rda` and `data/ord.fcs.raw.meta.df.out_mpc.rda` 
# which were generated from 266 .FCS files from 266 wells with 50 cells in each file.

## MPC - generating example data (random 50 per file)
{
  inpath="/path-to-MPCfcs/"
  subsampath="/path-to-subsamp/"
  
  ## Subsampling
  input_events_downsampling=50
  extra_args_read_FCS=NULL
  
  files <- list.files(inpath,full.names=TRUE,recursive=TRUE,pattern=".fcs")
  invisible(
    lapply(
      files,
      function(file){
        res <- do.call(read.FCS,c(list(filename=file),extra_args_read_FCS))
        w <- sort(sample(seq_len(nrow(res)),min(input_events_downsampling,nrow(res))))
        res <- res[w,]
        write.FCS(res,file.path(subsampath, basename(file)))
      }
    )
  )
}

### FFC
# This mice splenocytes dataset contains 50 cells (sorted CD4+ and CD8+ T cells) in each .FCS files which 
# was down-sampled from the data provided by Jalal Alshaweesh (Oct 2023) on [FlowRepository](http://flowrepository.org/id/FR-FCM-Z6UG). 
# The raw protein intensities and the corresponding metadata were saved in
# `data/ord.fcs.raw.mt_ffc.rda` and `data/ord.fcs.raw.meta.df.out_ffc.rda`.

## FFC - generating example data (random 50 per file)
{
  inpath="/path-to-FFCfcs/"
  subsampath="/path-to-subsamp/"
  
  ## Subsampling
  input_events_downsampling=50
  extra_args_read_FCS=NULL
  
  files <- list.files(inpath,full.names=TRUE,recursive=TRUE,pattern=".fcs")
  invisible(
    lapply(
      files,
      function(file){
        res <- do.call(read.FCS,c(list(filename=file),extra_args_read_FCS))
        w <- sort(sample(seq_len(nrow(res)),min(input_events_downsampling,nrow(res))))
        res <- res[w,]
        write.FCS(res,file.path(subsampath, basename(file)))
      }
    )
  )
}

---
title: "MAPFX: MAssively Parallel Flow cytometry Xplorer"
author: "Hsiao-Chi Liao and Agus Salim"
date: "`r Sys.Date()`"
output:
  BiocStyle::html_document:
    toc: true
    toc_depth: 2
vignette: >
  %\VignetteIndexEntry{MAPFX: MAssively Parallel Flow cytometry Xplorer}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: ref.bib
link-citations: true
---

```{r, setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```

# Introduction
## Motivation
Massively-Parallel Cytometry (MPC) experiments allow cost-effective quantification of more than 200 surface proteins at single-cell resolution. The Inflow protocol [@inflow] is the pioneer of the pipeline for analysing MPC data, and the Bioconductor's `infinityFlow` package was developed for comprehensive analyses. However, the methods for background correction and removal of unwanted variation implemented in the package can be improved. We develop the `MAPFX` package as an alternative that has a more thoughtful strategy to clean up the raw protein intensities. Unique features of our package compared to the `infinityFlow` pipeline include performing background correction prior to imputation and removing unwanted variation from the data at the cell-level, while explicitly accounting for the potential association between biology and unwanted factors. We benchmarked our pipeline against the `infinityFlow` pipeline and demonstrated that our approach is better at preserving biological signals, removing unwanted variation, and imputing unmeasured infinity markers [@mapfx]. Two user friendly functions `MapfxMPC` and `MapfxFFC` are included in the `MAPFX` package that were designed for data from either MPC or FFC experiments (see below sections for details). 

## Experimental and Computational Pipeline of the Data from the Massively-Parallel Cytometry (MPC) Experiments

![](F1.exp_compu_pipeline.png)
The experimental and the computational pipeline of the Inflow protocol [@inflow]: (A) Experimental pipeline. The single-cell samples are stained with backbone markers (backbone panel staining), then the stained samples are allocated to wells with one particular infinity marker (infinity panel staining), lastly, data can be acquired from the flow cytometry assay for each well. (B) Computational pipeline. The matrix of the normalised data showing that the backbone matrix (gray) contains values for every single-cell (row), but only block diagonal entries of the infinity matrix (yellow) have measurements. Imputation of the unmeasured infinity markers is done by using the backbone markers as predictors in regression models. Finally, the completed data matrix is obtained after imputation. The above figure is extracted from Figure 1 of the paper by @mapfx.

## Analysing Data from MPC Experiments
This package implemented an end-to-end toolbox for analysing raw data from MPC experiments. More details on the methodology can be found in @mapfx. The `MapfxMPC` function is designed for running through the whole pipeline. The pipeline starts by performing background correction on raw intensities to remove the noise from electronic baseline restoration and fluorescence compensation by adapting a normal-exponential convolution model. Unwanted technical variation, from sources such as well effects, is then removed using a log-normal model with plate, column, and row factors, after which infinity markers are imputed using the informative backbone markers as predictors with machine learning models. Cluster analysis and visualisation with UMAP two-dimensional representations can then be carried out if desired. Users can set `MapfxMPC(..., impute=FALSE)` if the imputation is not needed. 

## Analysing Data from the Fluorescence Flow Cytometry (FFC) Experiments
For the protein intensities from FFC experiments, the function `MapfxFFC` is used to carry out normalisation steps which include background correction and removal of unwanted variation, and the function can further perform cluster analysis and visualisation with UMAP two-dimensional representations if specified.

## Preparing Data for the Analysis - the Folder Diagram
```
# FCSpath
└───FCSpath
│   └───fcs
│       │   Plate1_A01.fcs
│       │   Plate1_A02.fcs
│       │   ...
│   └───meta
│       │   filename_meta.csv

# Outpath
└───Outpath
│   └───intermediary
│   └───downstream
│   └───graph

## Note: the sub-folders `intermediary`, `downstream`, and `graph` will 
## be generated automatically by MAPFX.
```

### Notes on Metadata
#### For MPC (the plate-based) Experiments
When set `file_meta = "auto"` for `MapfxMPC`, the file identifier keyword (GUID) of the FCS files MUST contain the following information and in the specified format:\
Plate information: Plate1, Plate2, …, Plate9\
Well information: A1, A2, …, A12, B1, …, H1, …, H12\

When set `file_meta = "usr"`, prepare `filename_meta.csv` in the following format and save the CSV file under `FCSpath/meta/`.\
An example:
```{r, metaformat1, echo = FALSE, results = 'asis'}
library(knitr)
df <- data.frame(
    Filenam = c("p1_a12.fcs", "p2_d08.fcs", "p3_g1.fcs"), #should be the same as the GUID in the FCS files
    Plate = c("Plate1", "Plate2", "Plate3"),
    Well = c("A12", "D08", "G01"),
    Column = c("Col.12", "Col.08", "Col.01"),
    Row = c("Row.01", "Row.04", "Row.07"),
    Well.lab = c("P1_A12", "P2_D08", "P3_G01"))
kable(df)
```
Note that the "Filenam" column refers to the GUID (file name) of each FCS file in the `FCSpath/fcs/`.


#### For FFC Experiments from Different Batches
Prepare `filename_meta.csv` in the following format and save the CSV file in `FCSpath/meta/`.\
An example:
```{r, metaformat2, echo = FALSE, results = 'asis'}
library(knitr)
df <- data.frame(
    Filenam = c("090122.fcs", "070122.fcs", "010122.fcs"), #should be the same as the GUID in the FCS files
    Batch = c("Batch1", "Batch2", "Batch3"))
kable(df)
```


# Analysing Data with the MAPFX Package
## Installation
The MAPFX package can be installed using the code below.
```{r, install_mapfx, eval=FALSE}
if (!require("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install("MAPFX")
```

Along with the MAPFX package, we also load the following packages required for running functions in MAPFX.
```{r, dependency}
library(MAPFX)

## specify the package names
suppressPackageStartupMessages({
library(flowCore)
library(Biobase)
library(stringr)
library(uwot)
library(iCellR)
library(igraph)
library(ggplot2)
library(RColorBrewer)
library(Rfast)
library(ComplexHeatmap)
library(circlize)
library(glmnetUtils)
library(e1071)
library(xgboost)
library(parallel)
library(pbapply)
library(reshape2)
library(gtools)
library(utils)
library(stats)
library(cowplot)
})

```

## Using the Example Datasets in MAPFX Package for this Vignette
### MPC
This dataset is a subset of the single-cell murine lung data at steady state downloaded from [FlowRepository](https://flowrepository.org/id/FR-FCM-Z2LP) provided by Etienne Becht (Nov 2020). The raw protein intensities and the corresponding metadata were saved in the objects `ord.fcs.raw.mt_mpc` and `ord.fcs.raw.meta.df.out_mpc` which were generated from 266 .FCS files from 266 wells with 50 cells in each file.

### FFC
This mice splenocytes dataset contains 50 cells (sorted CD4+ and CD8+ T cells) in each .FCS files which was down-sampled from the data provided by Jalal Alshaweesh (Oct 2023) on [FlowRepository](http://flowrepository.org/id/FR-FCM-Z6UG). The raw protein intensities and the corresponding metadata were saved in the objects `ord.fcs.raw.mt_ffc` and `ord.fcs.raw.meta.df.out_ffc`.

## `MapfxMPC(..., impute=TRUE)` - analysing data from MPC experiments
For users who would like to perform all of the following steps: background correction, removal of unwanted variation (well effects), **imputation**, and cluster analysis.

```{r, MapfxMPC_impu_example, eval=TRUE}
# import built-in data
data(ord.fcs.raw.meta.df.out_mpc)
data(ord.fcs.raw.mt_mpc)

# create an Output directory in the current working directory for the argument 'Outpath' of the MapfxMPC function
dir.create(file.path(tempdir(), "MPC_impu_Output"))

# usage
# when impute = TRUE, randomly selecting 50% of the cells in each well for model training
set.seed(123) 
MapfxMPC_impu_obj <- MapfxMPC(
    runVignette = TRUE, #set FALSE if not running this Vignette
    runVignette_meta = ord.fcs.raw.meta.df.out_mpc, #set NULL if not running this Vignette
    runVignette_rawInten = ord.fcs.raw.mt_mpc, #set NULL if not running this Vignette
    FCSpath = NULL, # users specify their own input path
    Outpath = file.path(tempdir(), "MPC_impu_Output"), # or users specify their own output path
    file_meta = "auto",
    bkb.v = c(
    "FSC-H", "FSC-W", "SSC-H", "SSC-W", "CD69-CD301b", "MHCII", 
    "CD4", "CD44", "CD8", "CD11c", "CD11b", "F480", 
    "Ly6C", "Lineage", "CD45a488", "CD24", "CD103"),
    yvar = "Legend", 
    control.wells = c(
    "P1_A01", "P2_A01", "P3_A01",
    "P3_F04", "P3_F05", "P3_F06", "P3_F07", "P3_F08", 
    "P3_F09", "P3_F10", "P3_F11", "P3_F12",
    "P3_G01", "P3_G02"),
    bkb.upper.quantile = 0.9, 
    bkb.lower.quantile = 0.1, 
    bkb.min.quantile = 0.01,
    inf.lower.quantile = 0.1, 
    inf.min.quantile = 0.01, 
    plots.bkc.bkb = TRUE, plots.bkc.inf = TRUE, 
    plots.initM = TRUE,
    plots.rmWellEffect = TRUE,
    impute = TRUE,
    models.use = c("XGBoost"),
    extra_args_regression_params = list(list(nrounds = 1500, eta = 0.03)),
    prediction_events_downsampling = NULL,
    impu.training = FALSE,
    plots.imputation = TRUE,
    cluster.analysis.bkb = TRUE, plots.cluster.analysis.bkb = TRUE,
    cluster.analysis.all = TRUE, plots.cluster.analysis.all = TRUE,
    cores = 4L)
    
# check the details
help(MapfxMPC, package = "MAPFX")
```

All the output will be stored in `file.path(tempdir(), "MPC_impu_Output")` (users can specify their own path: `/Outpath/`).

## `MapfxMPC(..., impute=FALSE)` - normalising data from MPC experiments
For users who would like to perform the following steps: background correction, removal of unwanted variation (well effects), and cluster analysis using backbones only.
```{r, mapfxNorm_MPC, eval=FALSE}
# import built-in data
data(ord.fcs.raw.meta.df.out_mpc)
data(ord.fcs.raw.mt_mpc)

# create an Output directory in the current working directory for the argument 'Outpath' of the MapfxMPC function
dir.create(file.path(tempdir(), "MPC_NOimpu_Output"))

# usage
MapfxMPC_NOimpu_obj <- MapfxMPC(
    runVignette = TRUE, #set FALSE if not running this Vignette
    runVignette_meta = ord.fcs.raw.meta.df.out_mpc, #set NULL if not running this Vignette
    runVignette_rawInten = ord.fcs.raw.mt_mpc, #set NULL if not running this Vignette
    FCSpath = NULL, # users specify their own input path
    Outpath = file.path(tempdir(), "MPC_NOimpu_Output"), # or users specify their own output path
    file_meta="auto",
    bkb.v = c(
    "FSC-H", "FSC-W", "SSC-H", "SSC-W", "CD69-CD301b", "MHCII", 
    "CD4", "CD44", "CD8", "CD11c", "CD11b", "F480", 
    "Ly6C", "Lineage", "CD45a488", "CD24", "CD103"),
    yvar="Legend", 
    control.wells = c(
    "P1_A01", "P2_A01", "P3_A01",
    "P3_F04", "P3_F05", "P3_F06", "P3_F07", "P3_F08", 
    "P3_F09", "P3_F10", "P3_F11", "P3_F12",
    "P3_G01", "P3_G02"),
    bkb.upper.quantile = 0.9, 
    bkb.lower.quantile = 0.1, 
    bkb.min.quantile = 0.01,
    inf.lower.quantile = 0.1, 
    inf.min.quantile = 0.01, 
    plots.bkc.bkb = TRUE, plots.bkc.inf = TRUE, 
    plots.initM = TRUE,
    plots.rmWellEffect = TRUE,
    impute = FALSE,
    cluster.analysis.bkb = TRUE, plots.cluster.analysis.bkb = TRUE,
    cores = 4L)

# check the details
help(MapfxMPC, package = "MAPFX")
```

All the output will be stored in `file.path(tempdir(), "MPC_NOimpu_Output")` (users can specify their own path: `/Outpath/`).

## `MapfxFFC` - normalising data from FFC experiments
For users who would like to perform the following steps: background correction, removal of unwanted variation (batch effects), and cluster analysis.
```{r, mapfxNorm_FFC, eval=TRUE}
# import built-in data
data(ord.fcs.raw.meta.df.out_ffc)
data(ord.fcs.raw.mt_ffc)

# create an Output directory in the current working directory for the argument 'Outpath' of the MapfxFFC function
dir.create(file.path(tempdir(), "FFCnorm_Output"))

MapfxFFC_obj <- MapfxFFC(
    runVignette = TRUE, #set FALSE if not running this Vignette
    runVignette_meta = ord.fcs.raw.meta.df.out_ffc, #set NULL if not running this Vignette
    runVignette_rawInten = ord.fcs.raw.mt_ffc, #set NULL if not running this Vignette
    FCSpath = NULL, # users specify their own input path
    Outpath = file.path(tempdir(), "FFCnorm_Output"), # or users specify their own output path
    protein.v = c("CD3","CD4","CD8","CD45"),
    protein.upper.quantile = 0.9, 
    protein.lower.quantile = 0.1, 
    protein.min.quantile = 0.01,
    plots.bkc.protein = TRUE,
    plots.initM = TRUE,
    plots.rmBatchEffect = TRUE,
    cluster.analysis.protein = TRUE, plots.cluster.analysis.protein = TRUE)
    
# check the details
help(MapfxFFC, package = "MAPFX")
```

All the output will be stored in `file.path(tempdir(), "FFCnorm_Output")` (users can specify their own path: `/Outpath/`).


## Description of the output
Three folders will be automatically generated in the output folder.\
1. `intermediary`:\
Intermediary results will be saved in the `.rds` or `.RData` formats and will be stored here.\
2. `downstream`:\
Final results will be saved in the `.rds` format and will be stored here. The results include normalised backbone measurements (on both linear and log scale: `bkc.adj.bkb_linearScale_mt.rds` and `bkc.adj.bkb_logScale_mt.rds`), the completed dataset with imputed infinity (exploratory, PE) markers (`predictions.Rds`), UMAP coordinates derived from both normalised backbones (`ClusterAnalysis_umap_#bkb.rds`) and the completed dataset (`ClusterAnalysis_ImpuMtd_umap_#bkb.#impuPE.rds`), and metadata (`fcs_metadata_df.rds`) for cells including cluster labels derived from both normalised backbones and the completed data matrix.\
3. `graph`:\
Figures will be stored here, including **scatter plots** for comparing background corrected and raw protein intensities for each protein marker, **heatmaps** for presenting the biological and unwanted effects in the data before and after removal of unwanted variation with *mapfx.norm*, **boxplots** (for imputations from multiple models) and **a boxplot and a histogram** (for imputations from a single model) of R-sq values for visualising the accuracy of imputed infinity (exploratory, PE) markers, and **UMAP plots** for showing the cluster structure.

## Examples of the output figures
The MapfxData package (soon will be available) contains two example datasets that can be used for demonstration.\

1. MPC dataset:\
It is a subset of the single-cell murine lung data at steady state downloaded from [FlowRepository](https://flowrepository.org/id/FR-FCM-Z2LP) [@inflow]. The raw data contains 266 .FCS files from 266 wells with 1000 cells in each file.

2. FFC dataset:\
It contains 316,779 cells (sorted CD4+ and CD8+ T cells) from mice splenocytes that was downloaded from [FlowRepository](http://flowrepository.org/id/FR-FCM-Z6UG) provided by Jalal Alshaweesh (Oct 2023).\

### Background Correction 
This works on both MPC and FFC data.

![Figure 1](bkc.jpg)

The comparison of background corrected values (y-axis) and raw intensities (x-axis) for a backbone marker (left, blue) and an infinity marker (right, gold) with a 45 degree line representing x=y. Our approach aims to "calibrate" the raw protein intensities, especially the non-positive values, without distorting large values too much.

### Removal of unwanted (well/batch) variation
This works on both MPC and FFC data.

![Figure 2](ruv_mpc.jpg)

A. Maximum likelihood estimates of the unwanted (left) and the biological (right) effects estimated from the pre-adjusted data. B. Maximum likelihood estimates of the unwanted (left) and the biological (right) effects estimated from the post-adjusted data using *mapfx.norm*. Orange represents positive effects, whereas blue indicates negative effects. The heatmaps show the existence of unwanted (well) effects and biological effects in the pre-adjusted data, and *mapfx.norm* managed to remove the unwanted (well) effects from the data while preserving biological variation. We can also use *mapfx.norm* to remove batch effect from FFC data.


### Performance of imputation
This works on MPC data.

![Figure 3](Rsq_boxplot_1model_500x800.jpeg)

The histogram and boxplot of the R-sq values of infinity markers. Higher R-sq values represent better performance of imputation.

### Cluster analysis
This works on both MPC and FFC data.

![Figure 4](cluster.jpg)

In this example, we show the results from MPC data, the UMAP two dimensional representation of cells with clusters derived from the PhenoGraph algorithm using the *mapfx.norm* normalised backbone data only (left) and both the normalised backbone and the imputed infinity markers (right). Clusters are better refined when we derived them using the completed data matrix (right).


# Session Inflromation
```{r, sessinfo}
sessionInfo()
```

# References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/021.bkc_bkb.R
\name{bkc_bkb}
\alias{bkc_bkb}
\title{Background correction for the backbone markers}
\usage{
bkc_bkb(
  paths,
  bkb.v,
  MPC,
  bkb.upper.quantile = 0.9,
  bkb.lower.quantile = 0.1,
  bkb.min.quantile = 0.01,
  plots = TRUE
)
}
\arguments{
\item{paths}{a vector of characters of paths to store intput, intermediary results, outputs...}

\item{bkb.v}{a vector of the names of the backbone markers (MUST match to the names in the FCS file).}

\item{MPC}{if the data is from MPC experiments, set \code{MPC = TRUE}. Setting FALSE represents data from the fluorescence flow cytometry (FFC) assay.}

\item{bkb.upper.quantile}{the cut-off (default = 0.9) for selecting cells used for estimating the parameter of signal.}

\item{bkb.lower.quantile}{the cut-off (default = 0.1) for selecting cells used for estimating the parameters of noise.}

\item{bkb.min.quantile}{the cut-off (default = 0.01) for omitting the cells with the smallest values to minimise the impact of outliers.}

\item{plots}{logical; if TRUE (default), produce scatter plots for pre- and post- background adjusted backbone markers (calibrated values on y-axis and raw values on x-axis).}
}
\value{
Background noise corrected backbone markers and graphs if specified
}
\description{
This function has been designed to do background correction for the backbone markers by using normal-exponential convolution model.
}
\details{
Generating the calibrated measurements and save to medpara_bkc.bkb_no.bkcPhy_mt.rds file, and visualising the result with the scatter plots in the output directory.
}
\author{
Hsiao-Chi Liao and Agus Salim
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BuiltinDataDoc.R
\docType{data}
\name{ord.fcs.raw.meta.df.out_mpc}
\alias{ord.fcs.raw.meta.df.out_mpc}
\title{Metadata of the example FFC data}
\format{
a data.frame
}
\usage{
data(ord.fcs.raw.meta.df.out_mpc)
}
\description{
Metadata of the example FFC data
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000.MapfxFFC.R
\name{MapfxFFC}
\alias{MapfxFFC}
\title{Normalising data from the Fluorescence Flow Cytometry (FFC) Experiments with mapfx.norm}
\usage{
MapfxFFC(
  runVignette = FALSE,
  runVignette_meta = NULL,
  runVignette_rawInten = NULL,
  FCSpath = NULL,
  Outpath = NULL,
  protein.v = NULL,
  protein.upper.quantile = 0.9,
  protein.lower.quantile = 0.1,
  protein.min.quantile = 0.01,
  plots.bkc.protein = TRUE,
  plots.initM = TRUE,
  plots.rmBatchEffect = TRUE,
  cluster.analysis.protein = TRUE,
  plots.cluster.analysis.protein = TRUE
)
}
\arguments{
\item{runVignette}{logical; if FALSE (default), specify a path to \code{FCSpath} argument; TRUE for running Vignette using built-in data.}

\item{runVignette_meta}{the argument for the built-in metadata when running Vignette; NULL (default).}

\item{runVignette_rawInten}{the argument for the built-in raw intensities when running Vignette; NULL (default).}

\item{FCSpath}{path to the input directory where \code{filename_meta.csv} and FCS files are stored. \code{filename_meta.csv} should be saved under \verb{FCSpath/FCS/meta/} and FCS files should be saved under \verb{FCSpath/FCS/fcs/}. See Vignette for details.}

\item{Outpath}{path to the output directory where intermediate results and final results will be stored.}

\item{protein.v}{a vector of the names of the protein markers (MUST be the same as the names in the FCS files). For example, \code{protein.v = c("FSC-H","FSC-W","SSC-H","SSC-W","CD3","CD4","CD8","CD45")}.}

\item{protein.upper.quantile}{the cut-off (default = 0.9) for selecting cells used for estimating the parameter of signal for protein markers.}

\item{protein.lower.quantile}{the cut-off (default = 0.1) for selecting cells used for estimating the parameters of noise for protein markers.}

\item{protein.min.quantile}{the cut-off (default = 0.01) for omitting the cells with the smallest values to minimise the impact of outliers during estimation.}

\item{plots.bkc.protein}{logical; if TRUE (default), produce scatter plots for pre- and post- background adjusted protein markers (calibrated values on y-axis and raw values on x-axis).}

\item{plots.initM}{logical; if TRUE (default), produce an UMAP embedding plot to visualise the structure of the biological clusters used to form the initial M matrix for removal of batch effects.}

\item{plots.rmBatchEffect}{logical; if TRUE (default), produce heatmaps to visualise the unwanted (batch) effects and biological effects in the pre- and post- adjusted datasets.}

\item{cluster.analysis.protein}{logical; if TRUE (default), perform cluster analysis using normalised protein markers.}

\item{plots.cluster.analysis.protein}{logical; if TRUE (default), produce an UMAP embedding plot from the normalised protein markers to visualise the structure of the biological clusters.}
}
\value{
Normalised protein markers on log scale and metadata for cells
}
\description{
This function is used to normalise, including background correction and removal of batch effects, protein intensity data from FFC assays. The input data is in FCS format. The functions include data normalisation and cluster analysis.
}
\details{
In the output directory, this function produces the normalised protein measurements, cell group labels from the cluster analysis using normalised proteins, and graphs will be provided if specified.
}
\examples{
# import built-in data
data(ord.fcs.raw.meta.df.out_ffc)
data(ord.fcs.raw.mt_ffc)

# create an Output directory for the MapfxFFC function
dir.create(file.path(tempdir(), "FFCnorm_Output"))

MapfxFFC_obj <- MapfxFFC(
  runVignette = TRUE, #set FALSE if not running this example
  runVignette_meta = ord.fcs.raw.meta.df.out_ffc, #set NULL if not running this example
  runVignette_rawInten = ord.fcs.raw.mt_ffc, #set NULL if not running this example
  FCSpath = NULL, # users specify their own input path if not running this example
  Outpath = file.path(tempdir(), "FFCnorm_Output"),
  protein.v = c("CD3","CD4","CD8","CD45"),
  protein.upper.quantile = 0.9, 
  protein.lower.quantile = 0.1, 
  protein.min.quantile = 0.01,
  plots.bkc.protein = TRUE,
  plots.initM = TRUE,
  plots.rmBatchEffect = TRUE,
  cluster.analysis.protein = TRUE, plots.cluster.analysis.protein = TRUE)
  

}
\author{
Hsiao-Chi Liao, Agus Salim
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/070.cluster.analysis_bkbOnly.R
\name{cluster.analysis.bkbOnly}
\alias{cluster.analysis.bkbOnly}
\title{Cluster analysis with normalised backbone measurements}
\usage{
cluster.analysis.bkbOnly(paths, bkb.v, plots = TRUE)
}
\arguments{
\item{paths}{a vector of characters of paths to store intput, intermediary results, outputs...}

\item{bkb.v}{a vector of the names of the backbone markers (MUST match to the names in the FCS file).}

\item{plots}{logical; if TRUE (default), produce an UMAP embedding plot from the normalised backbone markers to visualise the structure of the biological clusters.}
}
\value{
Metadata for cells with group labels from the cluster analysis
}
\description{
This function has been designed to perform cluster analysis for the normalised backbone measurements.
}
\details{
Updating the metadata for cells in the fcs_metadata_df.rds file, adding the information of the biological clusters from the clean and complete dataset, and visualising the result with the scatter plots in the output directory.
}
\author{
Hsiao-Chi Liao
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/032.rmWellEffect.R
\name{rmWellEffect}
\alias{rmWellEffect}
\title{Removing well effect from the data}
\usage{
rmWellEffect(paths, plots = TRUE)
}
\arguments{
\item{paths}{a vector of characters of paths to store intput, intermediary results, outputs...}

\item{plots}{logical; if TRUE (default), produce heatmaps to visualise the unwanted (well) effects and biological effects in the pre- and post- adjusted datasets.}
}
\value{
Normalised backbone markers on log scale
}
\description{
This function has been designed to remove the unwanted effects (well effects) from the background corrected measurements.
}
\details{
Generating the calibrated measurements and saving to bkc.adj.bkb_logScale_mt.rds (on log scale) and bkc.adj.bkb_linearScale_mt.rds (on linear scale), and visualising the result with the heatmaps in the output directory.
}
\author{
Hsiao-Chi Liao
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/031.initM.R
\name{initM}
\alias{initM}
\title{Initial biological clusters}
\usage{
initM(paths, assay, bkb.v, plots = TRUE)
}
\arguments{
\item{paths}{a vector of characters of paths to store intput, intermediary results, outputs...}

\item{assay}{the type of the input data - MPC or FFC.}

\item{bkb.v}{a vector of the names of the backbone markers (MUST match to the names in the FCS file).}

\item{plots}{logical; if TRUE (default), produce a UMAP embedding plot to visualise the structure of the biological clusters to form the initial M matrix for removal of well effect.}
}
\value{
Metadata for cells with the initial biological clusters labels added
}
\description{
Generating the M matrix for removing well effects.
}
\details{
This function has been designed to find initial biological clusters with centred transformed data.

Updating the metadata for cells in the fcs_metadata_df.rds file, adding the information of the initial biological clusters, and visualising the result with the scatter plots in the output directory.
}
\author{
Hsiao-Chi Liao
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/011.fcs_to_rds.R
\name{fcs_to_rds}
\alias{fcs_to_rds}
\title{Converting FCS files to RDS files}
\usage{
fcs_to_rds(paths, file_meta, yvar)
}
\arguments{
\item{paths}{a vector of characters of paths to store intput, intermediary results, outputs...}

\item{file_meta}{if the file names of the FCS files are in the specified format, set file_meta="auto"; otherwise set file_meta="usr" and provide "filename_meta.csv" in FCSpath.}

\item{yvar}{the name of the well-specific marker in the FCS files (e.g., "Legend").}
}
\value{
Raw protein intensities and the corresponding metadata from MPC experiments
}
\description{
This function has been designed to convert the raw FCS files to data matrix and export to RDS files.
}
\details{
Generating fcs_metadata_df.rds and fcs_rawInten_mt.rds files in the output directory.
}
\author{
Hsiao-Chi Liao
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/022.bkc_pe.R
\name{bkc_pe}
\alias{bkc_pe}
\title{Background correction for the well-specific markers (PE)}
\usage{
bkc_pe(paths, pe.lower.quantile = 0.1, pe.min.quantile = 0.01, plots = TRUE)
}
\arguments{
\item{paths}{a vector of characters of paths to store intput, intermediary results, outputs...}

\item{pe.lower.quantile}{the cut-off (default = 0.1) for selecting cells used for estimating the parameters of noise for infinity markers.}

\item{pe.min.quantile}{the cut-off (default = 0.01) for omitting the cells with the smallest values to minimise the impact of outliers for infinity markers.}

\item{plots}{logical; if TRUE (default), produce scatter plots for pre- and post- background adjusted infinity markers (calibrated values on y-axis and raw values on x-axis).}
}
\value{
Background noise corrected infinity markers and graphs if specified
}
\description{
This function has been designed to do background correction for the well-specific markers (PE) by using normal-exponential convolution model.
}
\details{
Generating the calibrated measurements and save to bkc.pe_mt.rds file, and visualising the result with the scatter plots in the output directory.
}
\author{
Hsiao-Chi Liao and Agus Salim
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/012.fcs_to_rds_bkb.R
\name{fcs_to_rds_bkb}
\alias{fcs_to_rds_bkb}
\title{Converting FCS files to RDS files (for the case without exploratory markers)}
\usage{
fcs_to_rds_bkb(paths, file_meta, MPC)
}
\arguments{
\item{paths}{a vector of characters of paths to store intput, intermediary results, outputs...}

\item{file_meta}{if the file names of the FCS files are in the specified format, set file_meta="auto"; otherwise set file_meta="usr" and provide "filename_meta.csv" in FCSpath.}

\item{MPC}{if the data is from MPC experiments, set MPC = TRUE. Setting FALSE represents data from the fluorescence flow cytometry (FFC) assay.}
}
\value{
Raw protein intensities and the corresponding metadata from FFC experiments
}
\description{
This function has been designed to convert the raw FCS files to data matrix and export to RDS files.
}
\details{
Generating fcs_metadata_df.rds and fcs_rawInten_mt.rds files in the output directory.
}
\author{
Hsiao-Chi Liao
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BuiltinDataDoc.R
\docType{data}
\name{ord.fcs.raw.meta.df.out_ffc}
\alias{ord.fcs.raw.meta.df.out_ffc}
\title{Metadata of the example MPC data}
\format{
a data.frame
}
\usage{
data(ord.fcs.raw.meta.df.out_ffc)
}
\description{
Metadata of the example MPC data
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/033.rmBatchEffect.R
\name{rmBatchEffect}
\alias{rmBatchEffect}
\title{Removing batch effect from the data (FFC)}
\usage{
rmBatchEffect(paths, plots = TRUE)
}
\arguments{
\item{paths}{a vector of characters of paths to store intput, intermediary results, outputs...}

\item{plots}{logical; if TRUE (default), produce heatmaps to visualise the unwanted (batch) effects and biological effects in the pre- and post- adjusted datasets.}
}
\value{
Normalised markers on log scale
}
\description{
This function has been designed to remove the unwanted effects (batch effects) from the background corrected measurements.
}
\details{
Generating the calibrated measurements and saving to bkc.adj.bkb_logScale_mt.rds (on log scale) and bkc.adj.bkb_linearScale_mt.rds (on linear scale), and visualising the result with the heatmaps in the output directory.
}
\author{
Hsiao-Chi Liao
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/051.imputation_bkb.predictors.R
\name{imputation_bkb.predictors}
\alias{imputation_bkb.predictors}
\title{Imputing the unmeasured well-specific markers with regression models}
\usage{
imputation_bkb.predictors(
  paths,
  chans,
  yvar = "Legend",
  cores = 4L,
  models.use,
  extra_args_regression_params,
  prediction_events_downsampling = NULL,
  impu.training,
  plots = TRUE
)
}
\arguments{
\item{paths}{a vector of characters of paths to store intput, intermediary results, outputs...}

\item{chans}{a vector of the names of the backbone markers (MUST match to the names in the FCS file).}

\item{yvar}{the name of the well-specific marker in the FCS files (has been changed to "Legend" in the first function).}

\item{cores}{the number of cores used to perform parallel computation (default = 8L).}

\item{models.use}{a vector of the names of the models used for imputation (an example: \code{c("LM", "LASSO3", "SVM", "XGBoost")}). The length of the vector is arbitrary.}

\item{extra_args_regression_params}{a list of the lists of the parameters for running regression models.}

\item{prediction_events_downsampling}{default = NULL (not doing subsampling). How many cells per well you want to have the imputation? (must be less than or equal to a half as we won't get the prediction for cells in the training set).}

\item{impu.training}{logical; if FALSE (default), not impute the training set (the dataset used to train the imputation models).}

\item{plots}{logical; if TRUE (default), visualise the distribution of R-sq from each infinity marker.}
}
\value{
A list of imputations
}
\description{
This function has been designed to impute/predict the unmeasured well-specific markers with regression models.
}
\details{
This function returns the object of imputation of the unmeasured well-specific markers. In the output directory, the imputations are saved to predictions.Rds file. Visualisation of the imputation accuracy will be provided if specified.
}
\author{
Hsiao-Chi Liao and InfinityFlow (Becht et. al, 2021)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BuiltinDataDoc.R
\docType{data}
\name{ord.fcs.raw.mt_mpc}
\alias{ord.fcs.raw.mt_mpc}
\title{Subset of the sorted CD4+ and CD8+ T cells from mice splenocytes from an FFC experiment}
\format{
a matrix containing cells from 5 batches (50 cells/batch)
}
\source{
\url{http://flowrepository.org/id/FR-FCM-Z6UG}
}
\usage{
data(ord.fcs.raw.mt_mpc)
}
\description{
Subset of the sorted CD4+ and CD8+ T cells from mice splenocytes from an FFC experiment
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BuiltinDataDoc.R
\docType{data}
\name{ord.fcs.raw.mt_ffc}
\alias{ord.fcs.raw.mt_ffc}
\title{Subset of the single-cell murine lung data at steady state from an MPC experiment}
\format{
a matrix containing cells from 266 wells (50 cells/well)
}
\source{
\url{https://flowrepository.org/id/FR-FCM-Z2LP}
}
\usage{
data(ord.fcs.raw.mt_ffc)
}
\description{
Subset of the single-cell murine lung data at steady state from an MPC experiment
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/060.cluster.analysis.R
\name{cluster.analysis}
\alias{cluster.analysis}
\title{Cluster analysis with normalised backbone measurements and the complete dataset}
\usage{
cluster.analysis(paths, bkb.v, yvar = "Legend", control.wells, plots = TRUE)
}
\arguments{
\item{paths}{a vector of characters of paths to store intput, intermediary results, outputs...}

\item{bkb.v}{a vector of the names of the backbone markers (MUST match to the names in the FCS file).}

\item{yvar}{the name of the well-specific marker in the FCS files (e.g., "Legend").}

\item{control.wells}{the well label of the control wells, including the autofluorescence and the isotype controls (format: plate_well, e.g., P1_A01)}

\item{plots}{logical; if TRUE (default), produce an UMAP embedding plot from the normalised backbone markers and the imputed infinity markers to visualise the structure of the biological clusters.}
}
\value{
Metadata for cells with group labels from the cluster analysis
}
\description{
This function has been designed to perform cluster analysis for the normalised backbone measurements and the complete dataset which includes the normalised backbone measurements and the imputed well-specific markers.
}
\details{
Updating the metadata for cells in the fcs_metadata_df.rds file, adding the information of the biological clusters from the clean and complete dataset, and visualising the result with the scatter plots in the output directory.
}
\author{
Hsiao-Chi Liao
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/040.mkImputeMT.R
\name{mkImputeMT}
\alias{mkImputeMT}
\title{Making the input for imputation}
\usage{
mkImputeMT(paths)
}
\arguments{
\item{paths}{a vector of characters of paths to store intput, intermediary results, outputs...}
}
\value{
Combined normalised backbone and infinity markers for imputation
}
\description{
This function has been designed to combine the normalised backbone measurements and the normalised PE markers for later imputation.
}
\details{
Generating the combined data and saving to impu.input_log.mt.rds (on log scale) in the output directory.
}
\author{
Hsiao-Chi Liao
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000.MapfxMPC.R
\name{MapfxMPC}
\alias{MapfxMPC}
\title{MAssively Parallel Flow cytometry Xplorer (MAPFX)}
\usage{
MapfxMPC(
  runVignette = FALSE,
  runVignette_meta = NULL,
  runVignette_rawInten = NULL,
  FCSpath = NULL,
  Outpath = NULL,
  file_meta = "auto",
  bkb.v = NULL,
  yvar = "Legend",
  control.wells = NULL,
  bkb.upper.quantile = 0.9,
  bkb.lower.quantile = 0.1,
  bkb.min.quantile = 0.01,
  inf.lower.quantile = 0.1,
  inf.min.quantile = 0.01,
  plots.bkc.bkb = TRUE,
  plots.bkc.inf = TRUE,
  plots.initM = TRUE,
  plots.rmWellEffect = TRUE,
  impute = TRUE,
  models.use = c("XGBoost"),
  extra_args_regression_params = list(list(nrounds = 1500, eta = 0.03)),
  prediction_events_downsampling = NULL,
  impu.training = FALSE,
  plots.imputation = TRUE,
  cluster.analysis.bkb = TRUE,
  plots.cluster.analysis.bkb = TRUE,
  cluster.analysis.all = TRUE,
  plots.cluster.analysis.all = TRUE,
  cores = 4L
)
}
\arguments{
\item{runVignette}{logical; if FALSE (default), specify a path to \code{FCSpath} argument; TRUE for running Vignette using built-in data.}

\item{runVignette_meta}{the argument for the built-in metadata when running Vignette; NULL (default).}

\item{runVignette_rawInten}{the argument for the built-in raw intensities when running Vignette; NULL (default).}

\item{FCSpath}{path to the input directory where \code{filename_meta.csv} and FCS files are stored (one file per well). \code{filename_meta.csv} should be saved under \verb{FCSpath/FCS/meta/} and FCS files should be saved under \verb{FCSpath/FCS/fcs/} (See Vignette for details.)}

\item{Outpath}{path to the output directory where intermediate results and final results will be stored.}

\item{file_meta}{if the file names of the FCS files are in the specified format, set \code{file_meta = "auto"}; otherwise set \code{file_meta = "usr"} and provide a \code{filename_meta.csv} file in \verb{FCSpath/FCS/meta/}.}

\item{bkb.v}{a vector of the names of the backbone markers (MUST be the same as the names in the FCS files). For example, \code{bkb.v = c("FSC-H", "FSC-W", "SSC-H", "SSC-W", "CD69-CD301b", "MHCII", "CD4", "CD44", "CD8", "CD11c", "CD11b", "F480", "Ly6C", "Lineage", "CD45a488", "CD24", "CD103")}.}

\item{yvar}{the name of the well-specific exploratory marker in the FCS files (e.g., "Legend").}

\item{control.wells}{the well label of the control wells, including the autofluorescence and the isotype controls (format: plate_well, e.g., P1_A01). Users need to provide this information when \code{cluster.analysis.all = TRUE}. For example, \code{control.wells = c("P1_A01", "P2_A01", "P3_A01", "P3_F04", "P3_F05", "P3_F06", "P3_F07", "P3_F08", "P3_F09", "P3_F10", "P3_F11", "P3_F12", "P3_G01", "P3_G02")}.}

\item{bkb.upper.quantile}{the cut-off (default = 0.9) for selecting cells used for estimating the parameter of signal for backbone markers.}

\item{bkb.lower.quantile}{the cut-off (default = 0.1) for selecting cells used for estimating the parameters of noise for backbone markers.}

\item{bkb.min.quantile}{the cut-off (default = 0.01) for omitting the cells with the smallest values to minimise the impact of outliers during estimation (backbone).}

\item{inf.lower.quantile}{the cut-off (default = 0.1) for selecting cells used for estimating the parameters of noise for infinity markers.}

\item{inf.min.quantile}{the cut-off (default = 0.01) for omitting the cells with the smallest values to minimise the impact of outliers during estimation (infinity).}

\item{plots.bkc.bkb}{logical; if TRUE (default), produce scatter plots for pre- and post- background adjusted backbone markers (calibrated values on y-axis and raw values on x-axis).}

\item{plots.bkc.inf}{logical; if TRUE (default), produce scatter plots for pre- and post- background adjusted infinity markers (calibrated values on y-axis and raw values on x-axis).}

\item{plots.initM}{logical; if TRUE (default), produce an UMAP embedding plot to visualise the structure of the biological clusters used to form the initial M matrix for removal of well effects.}

\item{plots.rmWellEffect}{logical; if TRUE (default), produce heatmaps to visualise the unwanted (well) effects and biological effects in the pre- and post- adjusted datasets.}

\item{impute}{logical; if TRUE (default), impute the missing infinity markers.}

\item{models.use}{a vector of the names of the models used for imputation. For example, \code{models.use = c("LM", "LASSO3", "SVM", "XGBoost")}.}

\item{extra_args_regression_params}{a list of the lists of the parameters for running regression models. The order should be the same as the models specified in \code{models.use}. For example, \code{extra_args_regression_params = list(list(degree = 1), list(nfolds = 10, degree = 3), list(type = "nu-regression", cost = 8, nu = 0.5, kernel = "radial"), list(nrounds = 1500, eta = 0.03))}.}

\item{prediction_events_downsampling}{integer (default = NULL); the number of samples used for the downsampling for the prediction.}

\item{impu.training}{logical; if FALSE (default), not impute the training set (the dataset used to train the imputation models).}

\item{plots.imputation}{logical; if TRUE (default), visualise the distribution of R-sq values of infinity markers.}

\item{cluster.analysis.bkb}{logical; if TRUE (default), perform cluster analysis using normalised backbone markers for all cells.}

\item{plots.cluster.analysis.bkb}{logical; if TRUE (default), produce an UMAP embedding plot from the normalised backbone markers to visualise the structure of the biological clusters for all cells.}

\item{cluster.analysis.all}{logical; must set \code{FALSE} if \code{impute = FALSE}; if TRUE (default), perform cluster analysis using normalised backbone markers and imputed infinity markers for cells in testing set.}

\item{plots.cluster.analysis.all}{logical; must set \code{FALSE} if \code{impute = FALSE}; if TRUE (default), produce an UMAP embedding plot from the normalised backbone markers and the imputed infinity markers to visualise the structure of the biological clusters for cells in testing set.}

\item{cores}{the number of cores used to perform parallel computation during the imputation process (default = 4L).}
}
\value{
Normalised backbone markers on log scale, background noise corrected infinity markers, imputations, and metadata for cells
}
\description{
This function is an end-to-end toolbox for analysing single-cell protein intensity data from the Massively-Parallel Cytometry (MPC) Experiments in FCS format. The functions include data normalisation, imputation (using backbone markers), and cluster analysis.
}
\details{
In the output directory, this function produces the normalised backbone measurements, the background corrected infinity measurements, and imputed infinity markers (if set impute = TRUE), cell group labels from the cluster analysis using both normalised backbones and the completed dataset (if impute = TRUE), and graphs will be provided if specified.
}
\examples{
# import built-in data
data(ord.fcs.raw.meta.df.out_mpc)
data(ord.fcs.raw.mt_mpc)

# create an Output directory for the MapfxMPC function
dir.create(file.path(tempdir(), "MPC_impu_Output"))

# When `impute = TRUE`, randomly selecting 50\% of the cells in each well for model training
set.seed(123) 
MapfxMPC_impu_obj <- MapfxMPC(
  runVignette = TRUE, #set FALSE if not running this example
  runVignette_meta = ord.fcs.raw.meta.df.out_mpc, #set NULL if not running this example
  runVignette_rawInten = ord.fcs.raw.mt_mpc, #set NULL if not running this example
  FCSpath = NULL, # users specify their own input path if not running this example
  Outpath = file.path(tempdir(), "MPC_impu_Output"),
  file_meta = "auto",
  bkb.v = c(
    "FSC-H", "FSC-W", "SSC-H", "SSC-W", "CD69-CD301b", "MHCII", 
    "CD4", "CD44", "CD8", "CD11c", "CD11b", "F480", 
    "Ly6C", "Lineage", "CD45a488", "CD24", "CD103"),
  yvar = "Legend", 
  control.wells = c(
    "P1_A01", "P2_A01", "P3_A01",
    "P3_F04", "P3_F05", "P3_F06", "P3_F07", "P3_F08", 
    "P3_F09", "P3_F10", "P3_F11", "P3_F12",
    "P3_G01", "P3_G02"),
  bkb.upper.quantile = 0.9, 
  bkb.lower.quantile = 0.1, 
  bkb.min.quantile = 0.01,
  inf.lower.quantile = 0.1, 
  inf.min.quantile = 0.01, 
  plots.bkc.bkb = TRUE, plots.bkc.inf = TRUE, 
  plots.initM = TRUE,
  plots.rmWellEffect = TRUE,
  impute = TRUE,
  models.use = c("XGBoost"),
  extra_args_regression_params = list(list(nrounds = 1500, eta = 0.03)),
  prediction_events_downsampling = NULL,
  impu.training = FALSE,
  plots.imputation = TRUE,
  cluster.analysis.bkb = TRUE, plots.cluster.analysis.bkb = TRUE,
  cluster.analysis.all = TRUE, plots.cluster.analysis.all = TRUE,
  cores = 2L)


}
\author{
Hsiao-Chi Liao, Agus Salim, and InfinityFlow (Becht et. al, 2021)
}
